import numpy as np
import pytest
from sparse import COO

from pyrovelocity.io.sparsity import (
    analyze_sparsification,
    calculate_density,
    check_sparsification_candidates,
    densify_arrays,
    print_sparsification_results,
    sparsify_arrays,
)


@pytest.fixture
def dense_array():
    return np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])


@pytest.fixture
def sparse_array():
    return np.array([[1, 0, 0], [0, 2, 0], [0, 0, 3]])


@pytest.fixture
def very_sparse_array():
    return np.array([[0, 0, 1], [0, 0, 0], [0, 2, 0]])


@pytest.fixture
def sample_dict_arrays(dense_array, sparse_array, very_sparse_array):
    return {
        "dense": dense_array,
        "sparse": sparse_array,
        "very_sparse": very_sparse_array,
    }


@pytest.fixture
def edge_cases_dict():
    return {
        "empty": np.array([]),
        "all_zeros": np.zeros((3, 3)),
        "single_element": np.array([1]),
        "single_zero": np.array([0]),
    }


@pytest.fixture
def large_sparse_array():
    arr = np.zeros((1000, 1000))
    arr[0, 0] = 1
    return arr


@pytest.mark.parametrize(
    "array,expected_density",
    [
        ("dense_array", 1.0),
        ("sparse_array", 1 / 3),
    ],
)
def test_calculate_density(array, expected_density, request):
    assert calculate_density(request.getfixturevalue(array)) == pytest.approx(
        expected_density
    )


def test_check_sparsification_candidates(sample_dict_arrays):
    candidates = check_sparsification_candidates(
        sample_dict_arrays, density_threshold=0.5
    )
    assert "dense" not in candidates
    assert "sparse" in candidates
    assert "very_sparse" in candidates


def test_print_sparsification_results(capsys, sample_dict_arrays):
    candidates = check_sparsification_candidates(
        sample_dict_arrays, density_threshold=0.5
    )
    print_sparsification_results(candidates)
    captured = capsys.readouterr()
    assert "Array 'sparse':" in captured.out
    assert "Array 'very_sparse':" in captured.out
    assert "Array 'dense':" not in captured.out


def test_analyze_sparsification(capsys, sample_dict_arrays):
    candidates = analyze_sparsification(
        sample_dict_arrays, density_threshold=0.5
    )
    assert "dense" not in candidates
    assert "sparse" in candidates
    assert "very_sparse" in candidates
    captured = capsys.readouterr()
    assert "Array 'sparse':" in captured.out
    assert "Array 'very_sparse':" in captured.out
    assert "Array 'dense':" not in captured.out


def test_sparsification_size_reduction(sample_dict_arrays):
    candidates = check_sparsification_candidates(
        sample_dict_arrays, density_threshold=0.5
    )
    for info in candidates.values():
        if info["density"] < 0.5:
            assert info["size_reduction"] >= 0.0


def test_edge_cases(edge_cases_dict):
    candidates = check_sparsification_candidates(
        edge_cases_dict, density_threshold=0.5
    )
    assert "empty" in candidates
    assert "all_zeros" in candidates
    assert "single_element" not in candidates
    assert "single_zero" in candidates


def test_large_array_performance(large_sparse_array):
    candidates = check_sparsification_candidates(
        {"large_sparse": large_sparse_array}, density_threshold=0.5
    )
    assert "large_sparse" in candidates
    assert candidates["large_sparse"]["density"] == pytest.approx(1 / 1_000_000)


@pytest.fixture
def mixed_dict_arrays(dense_array, sparse_array, very_sparse_array):
    return {
        "dense": dense_array,
        "sparse": sparse_array,
        "very_sparse": very_sparse_array,
        "already_coo": COO.from_numpy(np.array([[0, 1], [0, 0]])),
    }


def test_sparsify_arrays(mixed_dict_arrays):
    result = sparsify_arrays(mixed_dict_arrays, density_threshold=0.5)

    assert isinstance(result["dense"], np.ndarray)
    assert isinstance(result["sparse"], COO)
    assert isinstance(result["very_sparse"], COO)
    assert isinstance(result["already_coo"], COO)

    np.testing.assert_array_equal(result["dense"], mixed_dict_arrays["dense"])
    np.testing.assert_array_equal(
        result["sparse"].todense(), mixed_dict_arrays["sparse"]
    )
    np.testing.assert_array_equal(
        result["very_sparse"].todense(), mixed_dict_arrays["very_sparse"]
    )
    assert result["already_coo"] is mixed_dict_arrays["already_coo"]


@pytest.fixture
def coo_dict_arrays():
    return {
        "dense": np.array([[1, 2], [3, 4]]),
        "sparse": COO.from_numpy(np.array([[1, 0], [0, 2]])),
        "very_sparse": COO.from_numpy(np.array([[0, 1], [0, 0]])),
    }


def test_densify_arrays(coo_dict_arrays):
    result = densify_arrays(coo_dict_arrays)

    assert all(isinstance(arr, np.ndarray) for arr in result.values())
    np.testing.assert_array_equal(result["dense"], coo_dict_arrays["dense"])
    np.testing.assert_array_equal(
        result["sparse"], coo_dict_arrays["sparse"].todense()
    )
    np.testing.assert_array_equal(
        result["very_sparse"], coo_dict_arrays["very_sparse"].todense()
    )


def test_round_trip(sample_dict_arrays):
    sparsified = sparsify_arrays(sample_dict_arrays, density_threshold=0.5)
    densified = densify_arrays(sparsified)

    assert all(isinstance(arr, np.ndarray) for arr in densified.values())
    for key in sample_dict_arrays:
        np.testing.assert_array_equal(densified[key], sample_dict_arrays[key])



import os
import pickle
from os import PathLike
from pathlib import Path

import numpy as np
from beartype import beartype
from beartype.typing import Any, Dict
from sparse import COO
from zstandard import (
    ZstdCompressionParameters,
    ZstdCompressor,
    ZstdDecompressor,
)

from pyrovelocity.io.sparsity import densify_arrays, sparsify_arrays
from pyrovelocity.logging import configure_logging

__all__ = ["CompressedPickle"]

logger = configure_logging(__name__)


@beartype
def get_cpu_count() -> int:
    """
    Safely determine the number of CPUs in the system.
    Falls back to a default value if it can't be determined.
    """
    try:
        return os.cpu_count() or 1
    except NotImplementedError:
        return 1


@beartype
def get_optimal_thread_count(cpu_count: int) -> int:
    """
    Determine the optimal number of threads based on CPU count.
    """
    if cpu_count <= 2:
        return cpu_count
    elif cpu_count <= 8:
        return cpu_count - 1
    else:
        return cpu_count - 2


CPU_COUNT = get_cpu_count()
COMPRESSION_THREADS = get_optimal_thread_count(CPU_COUNT)


# TODO: Handle sparsification when values are not exclusively arrays
class CompressedPickle:
    """
    A class for reading and writing zstandard-compressed pickle files.

    Examples:
    >>> import pandas as pd
    >>> tmp = getfixture("tmp_path")
    >>> test_data_path = tmp / "test_data.pkl.zst"
    >>> test_data = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
    >>> CompressedPickle.save(test_data_path, test_data)
    >>> loaded_data = CompressedPickle.load(test_data_path)
    >>> loaded_data.equals(test_data)
    True
    """

    @staticmethod
    def save(
        file_path: PathLike | str,
        obj: Any,
        sparsify: bool = True,
    ) -> Path:
        """
        Save the given object to a zstandard-compressed pickle file.

        Args:
            file_path (str): The path of the file to save the object to.
            obj (object): The object to be saved.

        Examples:
        >>> import pandas as pd
        >>> tmp = getfixture("tmp_path")
        >>> test_data_path = tmp / "test_data.pkl.zst"
        >>> test_data = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
        >>> CompressedPickle.save(test_data_path, test_data)
        """
        file_path = Path(file_path)
        file_path.parent.mkdir(parents=True, exist_ok=True)

        if sparsify:
            if isinstance(obj, Dict) and all(
                isinstance(v, (np.ndarray, COO)) for v in obj.values()
            ):
                obj = sparsify_arrays(obj)
            else:
                logger.warning(
                    """
                    The object is not a dictionary of numpy arrays or COO objects.
                    It cannot be automatically sparsified.
                    """
                )

        compression_params = ZstdCompressionParameters(
            compression_level=9,
            threads=COMPRESSION_THREADS,
        )

        with file_path.open("wb") as f:
            compression_context = ZstdCompressor(
                compression_params=compression_params
            )
            with compression_context.stream_writer(f) as compressor:
                pickle.dump(obj, compressor)

        return file_path

    @staticmethod
    def load(
        file_path: PathLike | str,
        densify: bool = True,
    ) -> Any:
        """
        Load an object from a zstandard-compressed pickle file.

        Args:
            file_path (str): The path of the file to load the object from.

        Returns:
            object: The loaded object.

        Examples:
        >>> import pandas as pd
        >>> tmp = getfixture("tmp_path")
        >>> test_data_path = tmp / "test_data.pkl.zst"
        >>> test_data = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
        >>> CompressedPickle.save(test_data_path, test_data)
        >>> loaded_data = CompressedPickle.load(test_data_path)
        >>> loaded_data.equals(test_data)
        True
        """
        with open(file_path, "rb") as f:
            decompression_context = ZstdDecompressor()
            with decompression_context.stream_reader(f) as decompressor:
                obj = pickle.load(decompressor)

        if densify:
            if isinstance(obj, Dict) and all(
                isinstance(v, (np.ndarray, COO)) for v in obj.values()
            ):
                obj = densify_arrays(obj)
            else:
                logger.warning(
                    """
                    The object is not a dictionary of numpy arrays or COO objects.
                    It cannot be automatically densified.
                    """
                )
        return obj



from importlib.resources import files

import pytest

from pyrovelocity.io.serialization import load_anndata_from_json


@pytest.fixture
def adata_preprocessed():
    fixture_file_path = (
        files("pyrovelocity.tests.data") / "preprocessed_3_4.json"
    )
    return load_anndata_from_json(fixture_file_path)



from pyrovelocity.plots._count_histograms import (
    plot_spliced_unspliced_histogram,
)
from pyrovelocity.plots._dynamics import (
    plot_deterministic_simulation_phase_portrait,
    plot_deterministic_simulation_trajectories,
)
from pyrovelocity.plots._experimental import plot_t0_selection
from pyrovelocity.plots._genes import plot_gene_ranking
from pyrovelocity.plots._parameters import (
    plot_parameter_posterior_distributions,
)
from pyrovelocity.plots._predictive import (
    extrapolate_prediction_sample_predictive,
    posterior_curve,
)
from pyrovelocity.plots._rainbow import rainbowplot
from pyrovelocity.plots._summary import plot_gene_selection_summary
from pyrovelocity.plots._time import (
    plot_posterior_time,
    plot_shared_time_uncertainty,
)
from pyrovelocity.plots._uncertainty import (
    cluster_violin_plots,
    get_posterior_sample_angle_uncertainty,
    plot_state_uncertainty,
)
from pyrovelocity.plots._vector_fields import plot_vector_field_summary

__all__ = [
    "cluster_violin_plots",
    "extrapolate_prediction_sample_predictive",
    "get_posterior_sample_angle_uncertainty",
    "plot_deterministic_simulation_phase_portrait",
    "plot_deterministic_simulation_trajectories",
    "plot_gene_ranking",
    "plot_parameter_posterior_distributions",
    "plot_posterior_time",
    "plot_shared_time_uncertainty",
    "plot_spliced_unspliced_histogram",
    "plot_state_uncertainty",
    "plot_t0_selection",
    "posterior_curve",
    "rainbowplot",
    "plot_vector_field_summary",
    "plot_gene_selection_summary",
]



import pytest

from pyrovelocity.plots import plot_shared_time_uncertainty


@pytest.mark.slow
@pytest.mark.integration
def test_model2_plot_shared_time_uncertainty(
    postprocessed_model2_data,
    posterior_samples_model2,
    data_model2_reports_path,
):
    shared_time_plot = data_model2_reports_path / "shared_time_plot.pdf"
    plot_shared_time_uncertainty(
        adata=postprocessed_model2_data,
        posterior_samples=posterior_samples_model2,
        vector_field_basis="umap",
        shared_time_plot=shared_time_plot,
    )


@pytest.mark.slow
@pytest.mark.integration
def test_model1_plot_shared_time_uncertainty(
    postprocessed_model1_data,
    posterior_samples_model1,
    data_model1_reports_path,
):
    shared_time_plot = data_model1_reports_path / "shared_time_plot.pdf"
    plot_shared_time_uncertainty(
        adata=postprocessed_model1_data,
        posterior_samples=posterior_samples_model1,
        vector_field_basis="umap",
        shared_time_plot=shared_time_plot,
    )



from typing import Tuple

from beartype import beartype
from flytekit.configuration import Config
from flytekit.interaction.string_literals import literal_map_string_repr
from flytekit.remote.remote import FlyteRemote
from omegaconf import DictConfig, OmegaConf

from pyrovelocity.utils import print_config_tree

__all__ = ["get_remote_task_results"]


@beartype
def get_remote_task_results(
    execution_id: str,
    task_id: str,
    endpoint: str = "flyte.cluster.pyrovelocity.net",
    remote_uri_protocol: str = "flyte://v1",
    remote_project: str = "pyrovelocity",
    remote_domain: str = "development",
) -> Tuple[DictConfig, DictConfig]:
    """
    Get the inputs and outputs of a remote Flyte task.

    Args:
        execution_id (str): Flyte execution id.
        task_id (str): Flyte task id. Defaults to "f26c1pjy-0-dn1-0-dn6".
        endpoint (str): Cluster endpoint url. Defaults to "flyte.cluster.pyrovelocity.net".
        remote_uri_protocol (str): Flyte remote uri protocol. Defaults to "flyte://v1".
        remote_project (str): Flyte remote project. Defaults to "pyrovelocity".
        remote_domain (str): Flyte remote domain. Defaults to "development".

    Returns:
        Tuple[DictConfig, DictConfig]: Autocompletable inputs and outputs of
            the remote task execution.

    Example:
        >>> # xdoctest: +REQUIRES(env:NETWORK_ACCESS==true)
        >>> from pyrovelocity.io.cluster import get_remote_task_results
        ...
        >>> execution_id = "pyrovelocity-2021-09-01-16-00-00-000000-utc"
        >>> task_id = "f26c1pjy-0-dn1-0-dn6"
        ...
        >>> (
        ...     model1_postprocessing_inputs,
        ...     model1_postprocessing_outputs,
        ... ) = get_remote_task_results(
        ...     execution_id=execution_id,
        ...     task_id=task_id,
        ... )
    """
    remote = FlyteRemote(
        Config.for_endpoint(endpoint=endpoint),
    )
    execution_uri = (
        f"{remote_uri_protocol}/{remote_project}/{remote_domain}/{execution_id}"
    )
    task_uri = f"{execution_uri}/{task_id}"
    inputs = remote.get(f"{task_uri}/i")
    outputs = remote.get(f"{task_uri}/o")
    inputs_dict = literal_map_string_repr(inputs.literals)
    inputs_dictconfig = OmegaConf.create(inputs_dict)
    print_config_tree(inputs_dict)

    outputs_dict = literal_map_string_repr(outputs.literals)
    outputs_dictconfig = OmegaConf.create(outputs_dict)
    print_config_tree(outputs_dict)
    return inputs_dictconfig, outputs_dictconfig



import json
from pathlib import Path

import numpy as np
import pandas as pd
from anndata import AnnData
from anndata import Raw as AnnDataRaw
from beartype import beartype
from beartype.typing import Any, Dict
from scipy import sparse

from pyrovelocity.utils import ensure_numpy_array

__all__ = [
    "serialize_anndata",
    "deserialize_anndata",
    "save_anndata_to_json",
    "load_anndata_from_json",
    "create_sample_anndata",
]


@beartype
def serialize_anndata(adata: AnnData | AnnDataRaw) -> Dict[str, Any]:
    """
    Serialize an AnnData object to a dictionary.

    Args:
        adata: AnnData object to serialize

    Returns:
        Dictionary representation of the AnnData object
    """
    if isinstance(adata, AnnDataRaw):
        adata = adata.to_adata()

    serialized = {
        "shape": adata.shape,
        "obs": adata.obs.reset_index().to_dict(orient="list"),
        "var": adata.var.reset_index().to_dict(orient="list"),
    }

    if adata.X is not None:
        X = ensure_numpy_array(adata.X)
        serialized["X"] = X.tolist() if X.size > 0 else [[], []]

    if adata.layers:
        serialized["layers"] = {
            key: ensure_numpy_array(value)
            for key, value in adata.layers.items()
        }

    if adata.obsm:
        serialized["obsm"] = {
            key: ensure_numpy_array(value) for key, value in adata.obsm.items()
        }

    if adata.varm:
        serialized["varm"] = {
            key: ensure_numpy_array(value) for key, value in adata.varm.items()
        }

    if adata.obsp:
        serialized["obsp"] = {
            key: ensure_numpy_array(value) for key, value in adata.obsp.items()
        }

    if adata.varp:
        serialized["varp"] = {
            key: ensure_numpy_array(value) for key, value in adata.varp.items()
        }

    if adata.uns:
        serialized["uns"] = adata.uns

    if adata.raw is not None:
        serialized["raw"] = serialize_anndata(adata.raw)

    return serialized


@beartype
def deserialize_anndata(data: Dict[str, Any]) -> AnnData | AnnDataRaw:
    """
    Deserialize a dictionary to an AnnData object.

    Args:
        data: Dictionary representation of an AnnData object

    Returns:
        Reconstructed AnnData object
    """
    if not all(key in data for key in ["obs", "var", "shape"]):
        raise ValueError("Invalid data format: missing required keys")

    adata_dict = {
        "var": pd.DataFrame(data["var"]).set_index("index")
        if "index" in data["var"]
        else pd.DataFrame(index=range(data["shape"][1])),
        "obs": pd.DataFrame(data["obs"]).set_index("index")
        if "index" in data["obs"]
        else pd.DataFrame(index=range(data["shape"][0])),
    }

    if "X" in data:
        X = np.array(data["X"])
        if X.size == 0:
            X = X.reshape(data["shape"])
        adata_dict["X"] = X

    if "layers" in data:
        adata_dict["layers"] = {
            key: np.array(value) for key, value in data["layers"].items()
        }

    if "obsm" in data:
        adata_dict["obsm"] = {
            key: np.array(value) for key, value in data["obsm"].items()
        }

    if "varm" in data:
        adata_dict["varm"] = {
            key: np.array(value) for key, value in data["varm"].items()
        }

    if "obsp" in data:
        adata_dict["obsp"] = {
            key: np.array(value) for key, value in data["obsp"].items()
        }

    if "varp" in data:
        adata_dict["varp"] = {
            key: np.array(value) for key, value in data["varp"].items()
        }

    if "uns" in data:
        adata_dict["uns"] = data["uns"]

    adata = AnnData(**adata_dict)

    if "raw" in data:
        raw_data = deserialize_anndata(data["raw"])
        if isinstance(raw_data, AnnData):
            adata.raw = raw_data
        else:
            adata.raw = AnnData(X=raw_data.X, var=raw_data.var)

    return adata if "obs" in data else AnnDataRaw(adata)


@beartype
def save_anndata_to_json(
    adata: AnnData,
    filename: str | Path,
) -> None:
    """
    Save an AnnData object to a JSON file.

    Args:
        adata: AnnData object to save
        filename: Name of the JSON file to save to
    """
    adata_dict = serialize_anndata(adata)
    with open(filename, "w") as f:
        json.dump(adata_dict, f, indent=4, cls=NumpyEncoder)


@beartype
def load_anndata_from_json(filename: str | Path) -> AnnData:
    """
    Load an AnnData object from a JSON file.

    Args:
        filename: Name of the JSON file to load from

    Returns:
        Reconstructed AnnData object
    """
    with open(filename, "r") as f:
        adata_dict = json.load(f)
    return deserialize_anndata(adata_dict)


@beartype
def create_sample_anndata(M: int, N: int) -> AnnData:
    """
    Create a sample AnnData object.

    Args:
        M: Number of observations
        N: Number of variables

    Returns:
        Sample AnnData object
    """
    adata_dict = {
        "X": np.random.random((M, N)),
        "obs": pd.DataFrame(
            {"batch": np.random.choice(["a", "b"], M)},
            index=[f"cell{i:03d}" for i in range(M)],
        ),
        "var": pd.DataFrame(index=[f"gene{i:03d}" for i in range(N)]),
        "obsm": {
            "a": np.random.random((M, 100)),
        },
        "layers": {
            "a": np.random.random((M, N)),
        },
        "uns": {"sample_key": "sample_value"},
    }
    return AnnData(**adata_dict)


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, sparse.spmatrix):
            return obj.todense().tolist()
        if isinstance(obj, pd.DataFrame):
            return obj.to_dict(orient="list")
        if isinstance(obj, pd.Series):
            return obj.to_dict()
        return super().default(obj)


def numpy_to_python(obj):
    if isinstance(obj, dict):
        return {key: numpy_to_python(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [numpy_to_python(item) for item in obj]
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif sparse.issparse(obj):
        return obj.todense().tolist()
    elif isinstance(obj, pd.DataFrame):
        return obj.to_dict(orient="list")
    elif isinstance(obj, pd.Series):
        return obj.to_dict()
    else:
        return obj



import pytest

from pyrovelocity.plots import plot_parameter_posterior_distributions


@pytest.mark.slow
@pytest.mark.integration
def test_model2_plot_parameter_posterior_distributions(
    postprocessed_model2_data,
    posterior_samples_model2,
    putative_model2_marker_genes,
    data_model2_reports_path,
):
    plot_parameter_posterior_distributions(
        adata=postprocessed_model2_data,
        posterior_samples=posterior_samples_model2,
        geneset=putative_model2_marker_genes,
        parameter_uncertainty_plot=data_model2_reports_path
        / "parameter_uncertainties.pdf",
    )


@pytest.mark.slow
@pytest.mark.integration
def test_model1_plot_parameter_posterior_distributions(
    postprocessed_model1_data,
    posterior_samples_model1,
    putative_model1_marker_genes,
    data_model1_reports_path,
):
    plot_parameter_posterior_distributions(
        adata=postprocessed_model1_data,
        posterior_samples=posterior_samples_model1,
        geneset=putative_model1_marker_genes,
        parameter_uncertainty_plot=data_model1_reports_path
        / "parameter_uncertainties.pdf",
    )



import pytest
import scanpy as sc

from pyrovelocity.analysis.analyze import top_mae_genes
from pyrovelocity.io.compressedpickle import CompressedPickle
from pyrovelocity.tasks.data import download_dataset
from pyrovelocity.tasks.postprocess import postprocess_dataset
from pyrovelocity.tasks.preprocess import preprocess_dataset
from pyrovelocity.tasks.summarize import summarize_dataset
from pyrovelocity.tasks.train import train_dataset
from pyrovelocity.utils import generate_sample_data


@pytest.fixture
def default_sample_data():
    """Default sample data from `pyrovelocity.utils.generate_sample_data`.

    Returns:
        AnnData: AnnData object to use for testing.
    """
    return generate_sample_data(random_seed=98)


@pytest.fixture
def default_sample_data_file(default_sample_data, tmp_path):
    """Create a .h5ad file from the default sample data."""
    file_path = tmp_path / "default_sample_data.h5ad"
    default_sample_data.write(file_path)
    return file_path


def integration_fixture_scope(fixture_name, config):
    if config.getoption("--cached-integration-fixtures", None):
        return "session"
    return "session"


@pytest.fixture(scope=integration_fixture_scope)
def tmp_data_dir(tmp_path_factory):
    integration_tests_dir = tmp_path_factory.mktemp("integration")
    print(
        f"\nTemporary integration tests directory:\n\n",
        f"{integration_tests_dir}\n",
    )
    return integration_tests_dir


@pytest.fixture(scope=integration_fixture_scope)
def tmp_unit_reports_dir(tmp_data_dir):
    unit_reports_dir = tmp_data_dir / "unit_reports"
    unit_reports_dir.mkdir(parents=True, exist_ok=True)
    print(
        f"\nTemporary unit reports directory:\n\n",
        f"{unit_reports_dir}\n",
    )
    return unit_reports_dir


@pytest.fixture(scope=integration_fixture_scope)
def simulated_dataset_path(tmp_data_dir):
    return download_dataset(
        data_set_name="simulated",
        data_external_path=tmp_data_dir / "data/external",
        source="simulate",
        n_obs=100,
        n_vars=200,
    )


@pytest.fixture(scope=integration_fixture_scope)
def preprocess_dataset_output(simulated_dataset_path, tmp_data_dir):
    return preprocess_dataset(
        data_set_name="simulated",
        adata=simulated_dataset_path,
        data_processed_path=tmp_data_dir / "data/processed",
        reports_processed_path=tmp_data_dir / "reports/processed",
    )


@pytest.fixture(scope=integration_fixture_scope)
def train_dataset_output(preprocess_dataset_output, tmp_data_dir):
    _, preprocessed_dataset_path, _ = preprocess_dataset_output
    return train_dataset(
        adata=preprocessed_dataset_path,
        models_path=tmp_data_dir / "models",
        max_epochs=200,
    )


@pytest.fixture(scope=integration_fixture_scope)
def data_model2_reports_path(train_dataset_output, tmp_unit_reports_dir):
    unit_reports_data_model2_path = (
        tmp_unit_reports_dir / train_dataset_output[0]
    )
    unit_reports_data_model2_path.mkdir(parents=True, exist_ok=True)
    print(
        f"\nTemporary data model reports directory:\n\n",
        f"{unit_reports_data_model2_path}\n",
    )
    return unit_reports_data_model2_path


@pytest.fixture(scope=integration_fixture_scope)
def postprocess_dataset_output(train_dataset_output):
    return postprocess_dataset(
        *train_dataset_output[:6],
        vector_field_basis="umap",
        number_posterior_samples=4,
    )


@pytest.fixture(scope=integration_fixture_scope)
def summarize_dataset_output(
    train_dataset_output,
    postprocess_dataset_output,
    tmp_data_dir,
):
    return summarize_dataset(
        *train_dataset_output[:2],
        model_path=train_dataset_output[3],
        pyrovelocity_data_path=postprocess_dataset_output[0],
        postprocessed_data_path=postprocess_dataset_output[1],
        cell_state="leiden",
        vector_field_basis="umap",
        reports_path=tmp_data_dir / "reports",
    )


@pytest.fixture(scope=integration_fixture_scope)
def train_dataset_model1_output(preprocess_dataset_output, tmp_data_dir):
    _, preprocessed_dataset_path, _ = preprocess_dataset_output
    return train_dataset(
        adata=preprocessed_dataset_path,
        model_identifier="model1",
        models_path=tmp_data_dir / "models",
        guide_type="auto_t0_constraint",
        offset=False,
        max_epochs=200,
    )


@pytest.fixture(scope=integration_fixture_scope)
def data_model1_reports_path(train_dataset_model1_output, tmp_unit_reports_dir):
    unit_reports_data_model1_path = (
        tmp_unit_reports_dir / train_dataset_model1_output[0]
    )
    unit_reports_data_model1_path.mkdir(parents=True, exist_ok=True)
    print(
        f"\nTemporary data model 1 unit reports directory:\n\n",
        f"{unit_reports_data_model1_path}\n",
    )
    return unit_reports_data_model1_path


@pytest.fixture(scope=integration_fixture_scope)
def postprocess_dataset_model1_output(train_dataset_model1_output):
    return postprocess_dataset(
        *train_dataset_model1_output[:6],
        vector_field_basis="umap",
        number_posterior_samples=4,
    )


@pytest.fixture(scope=integration_fixture_scope)
def summarize_dataset_model1_output(
    train_dataset_model1_output,
    postprocess_dataset_model1_output,
    tmp_data_dir,
):
    return summarize_dataset(
        *train_dataset_model1_output[:2],
        model_path=train_dataset_model1_output[3],
        pyrovelocity_data_path=postprocess_dataset_model1_output[0],
        postprocessed_data_path=postprocess_dataset_model1_output[1],
        cell_state="leiden",
        vector_field_basis="umap",
        reports_path=tmp_data_dir / "reports",
    )


@pytest.fixture(scope=integration_fixture_scope)
def postprocessed_model2_data(postprocess_dataset_output):
    return sc.read(postprocess_dataset_output[1])


@pytest.fixture(scope=integration_fixture_scope)
def postprocessed_model1_data(postprocess_dataset_model1_output):
    return sc.read(postprocess_dataset_model1_output[1])


@pytest.fixture(scope=integration_fixture_scope)
def posterior_samples_model2(postprocess_dataset_output):
    return CompressedPickle.load(postprocess_dataset_output[0])


@pytest.fixture(scope=integration_fixture_scope)
def putative_model2_marker_genes(posterior_samples_model2):
    return top_mae_genes(
        volcano_data=posterior_samples_model2["gene_ranking"],
    )


@pytest.fixture(scope=integration_fixture_scope)
def posterior_samples_model1(postprocess_dataset_model1_output):
    return CompressedPickle.load(postprocess_dataset_model1_output[0])


@pytest.fixture(scope=integration_fixture_scope)
def putative_model1_marker_genes(posterior_samples_model1):
    return top_mae_genes(
        volcano_data=posterior_samples_model1["gene_ranking"],
    )



import os
import tarfile
from pathlib import Path

from beartype import beartype
from beartype.typing import List, Union
from fsspec import AbstractFileSystem, filesystem
from fsspec.implementations.local import LocalFileSystem
from returns.result import Failure, Result, Success

from pyrovelocity.logging import configure_logging

__all__ = [
    "create_tarball_from_filtered_dir",
    "copy_files_to_directory",
]

logger = configure_logging(__name__)


@beartype
def create_tarball_from_filtered_dir(
    src_dir: Path | str,
    output_filename: Path | str,
    extensions: tuple[str, ...] = (".png", ".pdf", ".csv"),
) -> Result[None, Exception]:
    """
    Creates a gzipped tarball of files with specific extensions under the given
    directory using fsspec.

    Args:
        src_dir: The top-level directory path to search for files.
        output_filename: The path to the output gzipped tarball file.
        extensions: A tuple of file extensions to include in the tarball. Defaults to ('.png', '.pdf').

    Returns:
        Result[None, Exception]: A result object encapsulating success or failure.

    Examples:
        >>> from pathlib import Path
        >>> tmp = getfixture('tmp_path')  # Get a temporary directory path
        >>> (tmp / 'document.pdf').write_text('PDF content')  # Create a PDF file
        >>> (tmp / 'image.png').write_text('PNG content')  # Create a PNG file
        >>> (tmp / 'exclude.txt').write_text('Should be excluded')  # Create a TXT file
        >>> output_tarball = tmp / 'output.tar.gz'  # Define the output tarball path
        >>> result = create_tarball_from_filtered_dir(tmp, output_tarball)  # Create tarball
        >>> print(f"Result: {result}") # Check the result
        Result: <Success: None>
        >>> import tarfile
        >>> with tarfile.open(output_tarball, 'r:gz') as tar:  # Open the created tarball
        ...     filenames = {member.name for member in tar.getmembers()}  # List contents
        >>> assert 'document.pdf' in filenames  # PDF should be included
        >>> assert 'image.png' in filenames  # PNG should be included
        >>> assert 'exclude.txt' not in filenames  # TXT should be excluded
    """
    try:
        logger.info(
            f"\nCreating gzipped tarball from {src_dir} for files with extensions:\n"
            f"\t{extensions}\n"
        )
        fs: AbstractFileSystem = LocalFileSystem()
        with tarfile.open(output_filename, "w:gz") as tar:
            for root, dirs, files in fs.walk(str(src_dir)):
                for file in files:
                    if file.endswith(extensions):
                        full_path = os.path.join(root, file)

                        # path is relative to the src_dir in the archive
                        relative_path = os.path.relpath(
                            full_path, start=str(src_dir)
                        )

                        with fs.open(full_path, "rb") as file_obj:
                            tarinfo = tarfile.TarInfo(name=relative_path)
                            tarinfo.size = file_obj.size
                            tar.addfile(tarinfo=tarinfo, fileobj=file_obj)
        return Success(None)

    except Exception as e:
        logger.error(
            f"Failed to create gzipped tarball from {src_dir} due to exception: {e}"
        )
        return Failure(e)


@beartype
def copy_files_to_directory(
    files_to_copy: List[str | Path],
    target_directory: str | Path,
) -> Result[None, Exception]:
    """
    Copy a list of files to a target directory using fsspec.

    Args:
        files_to_copy: A list of file paths (str or Path) to copy.
        target_directory: The target directory path (str or Path) to copy files into.

    Returns:
        Result[None, Exception]: A result object encapsulating success or failure.

    Examples:
        >>> from pathlib import Path
        >>> tmp = getfixture('tmp_path')  # Get a temporary directory path
        >>> source_dir = tmp / 'source'
        >>> source_dir.mkdir()  # Create a source directory
        >>> (source_dir / 'file1.txt').write_text('Content 1')  # Create a text file
        >>> (source_dir / 'file2.csv').write_text('Content 2')  # Create a CSV file
        >>> target_dir = tmp / 'target'
        >>> files_to_copy = [source_dir / 'file1.txt', source_dir / 'file2.csv']
        >>> result = copy_files_to_directory(files_to_copy, target_dir)  # Copy files
        >>> print(f"Result: {result}")  # Check the result
        Result: <Success: None>
        >>> assert (target_dir / 'file1.txt').exists()  # Check if file1.txt was copied
        >>> assert (target_dir / 'file2.csv').exists()  # Check if file2.csv was copied
        >>> with open(target_dir / 'file1.txt', 'r') as f:
        ...     content = f.read()
        ...     assert content == 'Content 1'  # Check if content was correctly copied
        >>> with open(target_dir / 'file2.csv', 'r') as f:
        ...     content = f.read()
        ...     assert content == 'Content 2'  # Check if content was correctly copied
    """
    fs: LocalFileSystem = filesystem("file")
    target_directory = str(target_directory)

    try:
        fs.makedirs(target_directory, exist_ok=True)

        for src_path in files_to_copy:
            src_path = str(src_path)
            filename = Path(src_path).name
            dst_path = str(Path(target_directory) / filename)

            try:
                with fs.open(src_path, "rb") as src_file, fs.open(
                    dst_path, "wb"
                ) as dst_file:
                    dst_file.write(src_file.read())
                print(f"Successfully copied {src_path} to {dst_path}")
            except Exception as e:
                print(f"Error copying {src_path} to {dst_path}: {e}")
                return Failure(e)

        return Success(None)

    except Exception as e:
        print(f"Unexpected error during file copying: {e}")
        return Failure(e)



import pytest

from pyrovelocity.plots import posterior_curve


@pytest.mark.slow
@pytest.mark.integration
def test_model2_plot_posterior_predictive(
    postprocessed_model2_data,
    posterior_samples_model2,
    putative_model2_marker_genes,
    train_dataset_output,
    data_model2_reports_path,
):
    posterior_curve(
        adata=postprocessed_model2_data,
        posterior_samples=posterior_samples_model2,
        gene_set=putative_model2_marker_genes,
        data_model=train_dataset_output[0],
        model_path=train_dataset_output[3],
        output_directory=data_model2_reports_path / "posterior_phase_portraits",
    )


@pytest.mark.slow
@pytest.mark.integration
def test_model1_plot_posterior_predictive(
    postprocessed_model1_data,
    posterior_samples_model1,
    putative_model1_marker_genes,
    train_dataset_model1_output,
    data_model1_reports_path,
):
    posterior_curve(
        adata=postprocessed_model1_data,
        posterior_samples=posterior_samples_model1,
        gene_set=putative_model1_marker_genes,
        data_model=train_dataset_model1_output[0],
        model_path=train_dataset_model1_output[3],
        output_directory=data_model1_reports_path / "posterior_phase_portraits",
    )



"""Tests for `pyrovelocity.tasks.postprocess` module."""

import pytest


def test_load_postprocess():
    from pyrovelocity.tasks import postprocess

    print(postprocess.__file__)


@pytest.mark.slow
@pytest.mark.integration
def test_postprocess_dataset(postprocess_dataset_output):
    return postprocess_dataset_output


@pytest.mark.slow
@pytest.mark.integration
def test_postprocess_dataset_model1(postprocess_dataset_model1_output):
    return postprocess_dataset_model1_output



from os import PathLike
from typing import Dict, List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from anndata import AnnData
from beartype import beartype
from matplotlib.figure import FigureBase

from pyrovelocity.logging import configure_logging

__all__ = ["plot_parameter_posterior_distributions"]

logger = configure_logging(__name__)


@beartype
def plot_parameter_posterior_distributions(
    posterior_samples: Dict[str, np.ndarray],
    adata: AnnData,
    geneset: List[str],
    parameter_uncertainty_plot: PathLike | str,
    parameter_names: List[str] = [
        "alpha",
        "beta",
        "gamma",
        "u_offset",
        "s_offset",
        "t0",
    ],
) -> FigureBase:
    parameters = [
        parameter
        for parameter in parameter_names
        if parameter in posterior_samples.keys()
    ]
    fig, ax = plt.subplots(len(parameters), 1)
    fig.set_size_inches(18, len(parameters) * 4)
    for index, parameter in enumerate(parameters):
        df = pd.DataFrame(
            np.log(
                posterior_samples[parameter].squeeze()[
                    :, np.isin(adata.var_names, list(geneset))
                ],
            ),
            columns=adata.var_names[np.isin(adata.var_names, list(geneset))],
        )
        # df = df.apply(lambda x: x - x.mean())
        df_long = df.melt(var_name="index", value_name="value")
        logger.debug(df_long.head())
        df_long["index"] = pd.Categorical(
            df_long["index"], categories=geneset, ordered=True
        )

        can_cast_all = (
            df_long["index"].astype(str).apply(lambda x: x.isdigit()).all()
        )
        if can_cast_all:
            df_long["index"] = (
                df_long["index"].astype(str).apply(lambda x: f"g_{int(x)}")
            )
            logger.warning(
                f"Converted integer index to string 'g_int' for {parameter}"
            )
        else:
            pass

        ax1 = sns.violinplot(
            x="index",
            y="value",
            data=df_long,
            ax=ax[index],
        )
        ax1.set_xticklabels(ax1.get_xticklabels(), rotation=30, ha="right")
        ax1.set_ylabel(parameter)
        ax1.set_xlabel("")

    fig.subplots_adjust(
        hspace=0.4, wspace=0.45, left=0.08, right=0.95, top=0.9, bottom=0.15
    )
    for ext in ["", ".png"]:
        fig.savefig(
            f"{parameter_uncertainty_plot}{ext}",
            facecolor=fig.get_facecolor(),
            bbox_inches="tight",
            edgecolor="none",
            dpi=300,
        )
    plt.close(fig)
    return fig



import altair as alt
import numpy as np
import polars as pl
from altair.vegalite.v5.api import Chart as AltairChart
from anndata import AnnData
from beartype import beartype
from scipy.sparse import csr_matrix

__all__ = ["plot_spliced_unspliced_histogram"]


alt.data_transformers.enable("vegafusion")


@beartype
def plot_spliced_unspliced_histogram(
    adata: AnnData,
    spliced_layer: str = "spliced",
    unspliced_layer: str = "unspliced",
    min_count: int = 3,
    max_count: int = 200,
    title: str = "",
    default_font_size: int = 24,
    plot_width: int = 700,
    plot_height: int = 700,
) -> AltairChart:
    spliced_matrix = csr_matrix(adata.layers[spliced_layer])
    unspliced_matrix = csr_matrix(adata.layers[unspliced_layer])

    non_zero = spliced_matrix.nonzero()

    spliced_values = spliced_matrix[non_zero].A1
    unspliced_values = unspliced_matrix[non_zero].A1

    df = pl.DataFrame(
        {"spliced": spliced_values, "unspliced": unspliced_values}
    )

    df = df.filter(
        (pl.col("spliced") >= min_count)
        & (pl.col("unspliced") >= min_count)
        & (pl.col("spliced") <= max_count)
        & (pl.col("unspliced") <= max_count)
    )

    number_of_histogram_bins = int(np.maximum(60, np.sqrt(len(df))))

    chart = (
        alt.Chart(df)
        .mark_rect()
        .encode(
            x=alt.X(
                "spliced:Q",
                title="Spliced Counts",
                bin=alt.Bin(maxbins=number_of_histogram_bins),
                axis=alt.Axis(tickMinStep=1),
            ),
            y=alt.Y(
                "unspliced:Q",
                title="Unspliced Counts",
                bin=alt.Bin(maxbins=number_of_histogram_bins),
                axis=alt.Axis(tickMinStep=1),
            ),
            color=alt.Color("count():Q", scale=alt.Scale(scheme="greenblue")),
            tooltip=["spliced", "unspliced", "count()"],
        )
        .properties(
            title=title,
            width=plot_width,
            height=plot_height,
        )
        .configure(
            countTitle="counts",
        )
        .configure_axis(
            labelFontSize=default_font_size,
            titleFontSize=default_font_size,
            titleFontWeight="normal",
        )
        .configure_title(fontSize=default_font_size)
        .configure_legend(
            orient="top-right",
            titleOrient="left",
            titleFontSize=default_font_size,
            titleFontWeight="normal",
            labelFontSize=default_font_size,
        )
        .interactive()
    )

    return chart



import pytest

from pyrovelocity.plots import plot_gene_ranking


@pytest.mark.slow
@pytest.mark.integration
def test_model2_plot_gene_ranking(
    postprocessed_model2_data,
    posterior_samples_model2,
    putative_model2_marker_genes,
    data_model2_reports_path,
):
    plot_gene_ranking(
        posterior_samples=[posterior_samples_model2],
        adata=[postprocessed_model2_data],
        selected_genes=putative_model2_marker_genes,
        time_correlation_with="st",
        show_marginal_histograms=True,
        save_volcano_plot=True,
        volcano_plot_path=data_model2_reports_path / "volcano.pdf",
    )


@pytest.mark.slow
@pytest.mark.integration
def test_model1_plot_gene_ranking(
    postprocessed_model1_data,
    posterior_samples_model1,
    putative_model1_marker_genes,
    data_model1_reports_path,
):
    plot_gene_ranking(
        posterior_samples=[posterior_samples_model1],
        adata=[postprocessed_model1_data],
        selected_genes=putative_model1_marker_genes,
        time_correlation_with="st",
        show_marginal_histograms=True,
        save_volcano_plot=True,
        volcano_plot_path=data_model1_reports_path / "volcano.pdf",
    )



"""Tests for `pyrovelocity.tasks.train` module."""

import pytest


def test_load_train():
    from pyrovelocity.tasks import train

    print(train.__file__)


@pytest.mark.slow
@pytest.mark.integration
def test_train_dataset(train_dataset_output):
    return train_dataset_output


@pytest.mark.slow
@pytest.mark.integration
def test_train_dataset_model1(train_dataset_model1_output):
    return train_dataset_model1_output



import json
import re
from collections import defaultdict
from datetime import timedelta
from pathlib import Path
from typing import Any, Dict

import pandas as pd
from beartype import beartype
from beartype.typing import Any, Dict, Tuple
from returns.result import Failure, Result, Success
from rich.console import Console
from rich.table import Table

__all__ = [
    "load_json",
    "merge_json",
    "combine_json_files",
    "add_duration_to_run_info",
    "generate_tables",
]


@beartype
def load_json(file_path: Path) -> Result[Dict[str, Any], Exception]:
    """
    Load JSON data from a file.

    Args:
        file_path: Path to the JSON file.

    Returns:
        Result[Dict[str, Any], Exception]: A result object containing the loaded JSON data or an exception.

    Examples:
        >>> import tempfile
        >>> tmp = getfixture('tmp_path')  # Get a temporary directory path
        >>> json_file = tmp / 'test.json'
        >>> json_file.write_text('{"key": "value"}')  # Create a test JSON file
        >>> result = load_json(json_file)
        >>> print(f"Result: {result}")
        Result: <Success: {'key': 'value'}>
        >>> assert result.unwrap() == {"key": "value"}
        >>> non_existent_file = tmp / 'non_existent.json'
        >>> result = load_json(non_existent_file)
        >>> print(f"Result: {result}")
        Result: <Failure: ...
    """
    try:
        with file_path.open("r") as file:
            return Success(json.load(file))
    except Exception as e:
        return Failure(e)


@beartype
def merge_json(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> Dict[str, Any]:
    """
    Merge two JSON dictionaries.

    Args:
        dict1: The first JSON dictionary.
        dict2: The second JSON dictionary.

    Returns:
        Dict[str, Any]: The merged dictionary.

    Examples:
        >>> dict1 = {"a": 1, "b": 2}
        >>> dict2 = {"b": 3, "c": 4}
        >>> result = merge_json(dict1, dict2)
        >>> print(result)
        {'a': 1, 'b': 3, 'c': 4}
        >>> assert result == {"a": 1, "b": 3, "c": 4}
    """
    return {**dict1, **dict2}


@beartype
def combine_json_files(
    file1: str | Path,
    file2: str | Path,
    output_file: str | Path,
) -> Result[None, Exception]:
    """
    Combine two JSON files and write the result to a new file.

    Args:
        file1: Path to the first JSON file.
        file2: Path to the second JSON file.
        output_file: Path to the output JSON file.

    Returns:
        Result[None, Exception]: A result object encapsulating success or failure.

    Examples:
        >>> tmp = getfixture('tmp_path')  # Get a temporary directory path
        >>> file1 = tmp / 'file1.json'
        >>> file2 = tmp / 'file2.json'
        >>> output_file = tmp / 'combined.json'
        >>> file1.write_text('{"a": 1, "b": 2}')
        >>> file2.write_text('{"b": 3, "c": 4}')
        >>> result = combine_json_files(file1, file2, output_file)
        >>> print(f"Result: {result}")
        Result: <Success: None>
        >>> assert output_file.exists()
        >>> with open(output_file, 'r') as f:
        ...     content = json.load(f)
        ...     assert content == {"a": 1, "b": 3, "c": 4}
    """
    try:
        data1_result = load_json(Path(file1))
        data2_result = load_json(Path(file2))

        if isinstance(data1_result, Failure):
            return data1_result
        if isinstance(data2_result, Failure):
            return data2_result

        data1 = data1_result.unwrap()
        data2 = data2_result.unwrap()

        combined_data = merge_json(data1, data2)

        with output_file.open("w") as file:
            json.dump(combined_data, file, indent=2)

        return Success(None)
    except Exception as e:
        return Failure(e)


@beartype
def add_duration_to_run_info(
    file_path: Path | str = "metrics.json"
) -> Result[None, Exception]:
    """
    Load a JSON file containing run information, add a duration field,
    and overwrite the file with the updated information.

    The duration is calculated as the difference between 'end_time' and 'start_time',
    and is added as a new field in the format "HH:MM:SS".

    Args:
        file_path (Path | str): Path to the JSON file. Defaults to "metrics.json".

    Returns:
        Result[None, Exception]: A Result object containing None if successful,
                                 or an Exception if an error occurred.

    Examples:
        >>> import tempfile
        >>> from pathlib import Path
        >>> # Create a temporary JSON file
        >>> with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.json') as temp_file:
        ...     json.dump({
        ...         "start_time": 1723744878851,
        ...         "end_time": 1723744887595,
        ...         "other_field": "value"
        ...     }, temp_file)
        ...     temp_file_path = Path(temp_file.name)
        >>> # Test the function
        >>> result = add_duration_to_run_info(temp_file_path)
        >>> print(f"Result: {result}")
        Result: <Success: None>
        >>> # Verify the file was updated correctly
        >>> with open(temp_file_path, 'r') as f:
        ...     updated_data = json.load(f)
        >>> print(f"Duration added: {updated_data['duration']}")
        Duration added: 00:00:08
        >>> assert 'duration' in updated_data
        >>> assert updated_data['duration'] == "00:00:08"
        >>> assert updated_data['other_field'] == "value"
        >>> # Clean up
        >>> temp_file_path.unlink()

        >>> # Test with non-existent file
        >>> result = add_duration_to_run_info("non_existent.json")
        >>> print(f"Result: {result}")
        Result: <Failure: ...

        >>> # Test with invalid JSON
        >>> with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.json') as temp_file:
        ...     temp_file.write("Invalid JSON")
        ...     temp_file_path = Path(temp_file.name)
        >>> result = add_duration_to_run_info(temp_file_path)
        >>> print(f"Result: {result}")
        Result: <Failure: ...
        >>> # Clean up
        >>> temp_file_path.unlink()
    """
    file_path = Path(file_path)

    try:
        with file_path.open("r") as file:
            data: Dict[str, Any] = json.load(file)

        start_time = data.get("start_time")
        end_time = data.get("end_time")

        if start_time is not None and end_time is not None:
            duration = timedelta(milliseconds=end_time - start_time)
            hours, remainder = divmod(duration.seconds, 3600)
            minutes, seconds = divmod(remainder, 60)

            data["duration"] = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
        else:
            return Failure(
                ValueError(
                    "Unable to calculate duration. Start time or end time is missing."
                )
            )

        with file_path.open("w") as file:
            json.dump(data, file, indent=2)

        return Success(None)

    except Exception as e:
        return Failure(e)


@beartype
def generate_tables(json_data: Dict[str, Any]) -> Tuple[str, str, str, Table]:
    """
    Generate LaTeX, HTML, and rich tables from the provided JSON data.

    Args:
        json_data (Dict[str, Any]):
            The JSON data containing metrics for different datasets and models.

    Returns:
        Tuple[str, str, Table]:
            A tuple containing the LaTeX table string, HTML table string,
            and rich Table object.

    Examples:
        >>> json_data = {
        ...     "simulated_model1-123": {"run_name": "simulated_model1-123", "-ELBO": -7.339, "MAE": 1.094},
        ...     "simulated_model2-456": {"run_name": "simulated_model2-456", "-ELBO": -7.512, "MAE": 1.123},
        ...     "pancreas_model1-789": {"run_name": "pancreas_model1-789", "-ELBO": -8.234, "MAE": 0.987},
        ...     "pancreas_model2-012": {"run_name": "pancreas_model2-012", "-ELBO": -8.123, "MAE": 0.998}
        ... }
        >>> latex, html, markdown, rich_table = generate_tables(json_data)
    """

    dataset_metrics = defaultdict(lambda: defaultdict(dict))
    for run_name, metrics in json_data.items():
        match = re.match(r"(\w+)_(model\d+)-", run_name)
        if match:
            dataset, model = match.groups()
            dataset_metrics[dataset][model]["-ELBO"] = metrics.get(
                "-ELBO", "N/A"
            )
            dataset_metrics[dataset][model]["MAE"] = metrics.get("MAE", "N/A")

    data = []
    for dataset, models in dataset_metrics.items():
        row = {"Dataset": dataset}
        for model in ["model1", "model2"]:
            if model in models:
                row[f"{model.capitalize()} -ELBO"] = models[model].get(
                    "-ELBO", "N/A"
                )
                row[f"{model.capitalize()} MAE"] = models[model].get(
                    "MAE", "N/A"
                )
            else:
                row[f"{model.capitalize()} -ELBO"] = "N/A"
                row[f"{model.capitalize()} MAE"] = "N/A"
        data.append(row)

    df = pd.DataFrame(data)

    df.columns = pd.MultiIndex.from_tuples(
        [
            ("Dataset", ""),
            ("Model 1", "-ELBO"),
            ("Model 1", "MAE"),
            ("Model 2", "-ELBO"),
            ("Model 2", "MAE"),
        ]
    )

    latex_table = df.to_latex(
        index=False,
        multicolumn=True,
        multicolumn_format="c",
        column_format="l|cc|cc",
    )
    latex_table = latex_table.replace("\\midrule", "\\midrule\n\\cline{2-5}")

    html_table = """
    <table border="1" class="dataframe">
      <thead>
        <tr>
          <th>Dataset</th>
          <th colspan="2">Model 1</th>
          <th colspan="2">Model 2</th>
        </tr>
        <tr>
          <th></th>
          <th>-ELBO</th>
          <th>MAE</th>
          <th>-ELBO</th>
          <th>MAE</th>
        </tr>
      </thead>
      <tbody>
        {rows}
      </tbody>
    </table>
    """

    rows = ""
    for _, row in df.iterrows():
        rows += f"""
        <tr>
          <td>{row[("Dataset", "")]}</td>
          <td>{row[("Model 1", "-ELBO")]:.4f}</td>
          <td>{row[("Model 1", "MAE")]:.4f}</td>
          <td>{row[("Model 2", "-ELBO")]:.4f}</td>
          <td>{row[("Model 2", "MAE")]:.4f}</td>
        </tr>
        """

    html_table = html_table.format(rows=rows)

    markdown_table = (
        "| Dataset   | Model 1   |          | Model 2   |          |\n"
    )
    markdown_table += (
        "|-----------|-----------|----------|-----------|----------|\n"
    )
    markdown_table += (
        "|           | -ELBO     | MAE      | -ELBO     | MAE      |\n"
    )
    for _, row in df.iterrows():
        markdown_table += (
            f"| {row['Dataset', '']} "
            f"| {row['Model 1', '-ELBO']:.4f} | {row['Model 1', 'MAE']:.4f} "
            f"| {row['Model 2', '-ELBO']:.4f} | {row['Model 2', 'MAE']:.4f} |\n"
        )

    rich_table = Table(title="Model Comparison")
    rich_table.add_column("Dataset", style="cyan", no_wrap=True)
    rich_table.add_column("Model 1", style="magenta", no_wrap=True)
    rich_table.add_column("", style="magenta", no_wrap=True)
    rich_table.add_column("Model 2", style="green", no_wrap=True)
    rich_table.add_column("", style="green", no_wrap=True)
    rich_table.add_row("", "-ELBO", "MAE", "-ELBO", "MAE")

    for _, row in df.iterrows():
        rich_table.add_row(
            row[("Dataset", "")],
            f"{row[('Model 1', '-ELBO')]:.4f}"
            if isinstance(row[("Model 1", "-ELBO")], float)
            else str(row[("Model 1", "-ELBO")]),
            f"{row[('Model 1', 'MAE')]:.4f}"
            if isinstance(row[("Model 1", "MAE")], float)
            else str(row[("Model 1", "MAE")]),
            f"{row[('Model 2', '-ELBO')]:.4f}"
            if isinstance(row[("Model 2", "-ELBO")], float)
            else str(row[("Model 2", "-ELBO")]),
            f"{row[('Model 2', 'MAE')]:.4f}"
            if isinstance(row[("Model 2", "MAE")], float)
            else str(row[("Model 2", "MAE")]),
        )

    console = Console()
    console.print(latex_table)
    console.print(html_table)
    console.print(markdown_table)
    console.print(rich_table)

    return (
        latex_table,
        html_table,
        markdown_table,
        rich_table,
    )



import json
from dataclasses import asdict
from datetime import timedelta
from pathlib import Path

from beartype.typing import List
from flytekit import Resources, current_context, dynamic, task, workflow
from flytekit.extras.accelerators import T4, GPUAccelerator
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile
from returns.result import Failure, Success

from pyrovelocity.interfaces import (
    DownloadDatasetInterface,
    PreprocessDataInterface,
    PyroVelocityTrainInterface,
)
from pyrovelocity.io.archive import (
    copy_files_to_directory,
    create_tarball_from_filtered_dir,
)
from pyrovelocity.io.gcs import upload_file_concurrently
from pyrovelocity.io.json import (
    add_duration_to_run_info,
    combine_json_files,
    generate_tables,
    load_json,
)
from pyrovelocity.logging import configure_logging
from pyrovelocity.tasks.data import download_dataset
from pyrovelocity.tasks.postprocess import postprocess_dataset
from pyrovelocity.tasks.preprocess import preprocess_dataset
from pyrovelocity.tasks.summarize import summarize_dataset
from pyrovelocity.tasks.train import train_dataset
from pyrovelocity.utils import str_to_bool
from pyrovelocity.workflows.constants import (
    PYROVELOCITY_CACHE_FLAG,
    PYROVELOCITY_DATA_SUBSET,
)
from pyrovelocity.workflows.main_configuration import (
    CombinedMetricsOutputs,
    PostprocessConfiguration,
    PostprocessOutputs,
    PreprocessOutputs,
    ResourcesJSON,
    SummarizeOutputs,
    TrainingOutputs,
    WorkflowConfiguration,
    bonemarrow_configuration,
    default_resource_limits,
    default_resource_requests,
    default_training_resource_limits,
    default_training_resource_requests,
    larry_configuration,
    larry_mono_configuration,
    larry_multilineage_configuration,
    larry_neu_configuration,
    pancreas_configuration,
    pbmc5k_configuration,
    pbmc10k_configuration,
    pbmc68k_configuration,
    pons_configuration,
    simulated_configuration,
)

__all__ = [
    "download_data",
    "preprocess_data",
    "train_model",
    "postprocess_data",
    "summarize_data",
    "upload_summary",
    "map_model_configurations_over_data_set",
    "training_workflow",
]

logger = configure_logging(__name__)

CACHE_VERSION = "2024.8.15"
DOWNLOAD_CACHE_VERSION = f"{CACHE_VERSION}.0"
PREPROCESS_CACHE_VERSION = f"{CACHE_VERSION}.0"
TRAIN_CACHE_VERSION = f"{CACHE_VERSION}.0"
POSTPROCESS_CACHE_VERSION = f"{CACHE_VERSION}.0"
SUMMARIZE_CACHE_VERSION = f"{CACHE_VERSION}.0"
UPLOAD_CACHE_VERSION = f"{CACHE_VERSION}.2"
COMBINE_METRICS_CACHE_VERSION = f"{CACHE_VERSION}.2"
L4 = GPUAccelerator("nvidia-l4")
ACCELERATOR_TYPE: GPUAccelerator = T4


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=DOWNLOAD_CACHE_VERSION,
    retries=3,
    interruptible=True,
    timeout=timedelta(minutes=20),
    requests=Resources(cpu="2", mem="4Gi", ephemeral_storage="16Gi"),
    limits=Resources(cpu="8", mem="16Gi", ephemeral_storage="200Gi"),
)
def download_data(download_dataset_args: DownloadDatasetInterface) -> FlyteFile:
    """
    Download external data.
    """
    dataset_path = download_dataset(**asdict(download_dataset_args))
    print(f"\nFlyte download data path: {dataset_path}\n")
    return FlyteFile(path=str(dataset_path))


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=PREPROCESS_CACHE_VERSION,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=60),
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="32Gi"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi"),
    enable_deck=False,
)
def preprocess_data(
    data: FlyteFile, preprocess_data_args: PreprocessDataInterface
) -> PreprocessOutputs:
    """
    Download external data.
    """
    data_path = data.download()
    print(f"\nFlyte preprocess input data path: {data_path}\n")
    preprocess_data_args.adata = str(data_path)
    _, processed_dataset_path, processed_reports_path = preprocess_dataset(
        **asdict(preprocess_data_args),
    )

    print(f"\nFlyte preprocess output data path: {processed_dataset_path}\n")
    return PreprocessOutputs(
        processed_data=FlyteFile(path=str(processed_dataset_path)),
        processed_reports=FlyteDirectory(path=str(processed_reports_path)),
    )


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=TRAIN_CACHE_VERSION,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=120),
    container_image="{{.image.gpu.fqn}}:{{.image.gpu.version}}",
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="50Gi", gpu="1"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi", gpu="1"),
    accelerator=ACCELERATOR_TYPE,
    enable_deck=False,
)
def train_model(
    preprocess_outputs: PreprocessOutputs,
    train_model_configuration: PyroVelocityTrainInterface,
) -> TrainingOutputs:
    """
    Train model.
    """
    processed_data_path = preprocess_outputs.processed_data.download()
    print(f"\nFlyte train model input data path: {processed_data_path}\n")
    train_model_configuration.adata = str(processed_data_path)
    (
        data_model,
        data_model_path,
        trained_data_path,
        model_path,
        posterior_samples_path,
        metrics_path,
        run_info_path,
        loss_plot_path,
        loss_csv_path,
    ) = train_dataset(
        **asdict(train_model_configuration),
    )

    print(
        f"\nFlyte train model outputs:\n"
        f"\tdata model: {data_model}\n"
        f"\tdata model path: {data_model_path}\n"
        f"\ttrained data path: {trained_data_path}\n"
        f"\tmodel path: {model_path}\n"
        f"\tposterior samples path: {posterior_samples_path}\n"
        f"\tmetrics path: {metrics_path}\n"
        f"\trun info path: {run_info_path}\n"
        f"\tloss plot path: {loss_plot_path}\n\n"
        f"\tloss csv path: {loss_csv_path}\n\n"
    )

    return TrainingOutputs(
        data_model=data_model,
        data_model_path=data_model_path,
        trained_data_path=FlyteFile(path=str(trained_data_path)),
        model_path=FlyteDirectory(path=str(model_path)),
        posterior_samples_path=FlyteFile(path=str(posterior_samples_path)),
        metrics_path=FlyteFile(path=str(metrics_path)),
        run_info_path=FlyteFile(path=str(run_info_path)),
        loss_plot_path=FlyteFile(path=str(loss_plot_path)),
        loss_csv_path=FlyteFile(path=str(loss_csv_path)),
    )


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=POSTPROCESS_CACHE_VERSION,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=120),
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="50Gi"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi"),
    enable_deck=False,
)
def postprocess_data(
    preprocess_data_args: PreprocessDataInterface,
    training_outputs: TrainingOutputs,
    postprocess_configuration: PostprocessConfiguration,
) -> PostprocessOutputs:
    trained_data_path = training_outputs.trained_data_path.download()
    model_path = training_outputs.model_path.download()
    posterior_samples_path = training_outputs.posterior_samples_path.download()
    metrics_path = training_outputs.metrics_path.download()

    pyrovelocity_data_path, postprocessed_data_path = postprocess_dataset(
        data_model=training_outputs.data_model,
        data_model_path=training_outputs.data_model_path,
        trained_data_path=trained_data_path,
        model_path=model_path,
        posterior_samples_path=posterior_samples_path,
        metrics_path=metrics_path,
        vector_field_basis=preprocess_data_args.vector_field_basis,
        number_posterior_samples=postprocess_configuration.number_posterior_samples,
    )

    print(
        f"\npostprocessed_data_path: {postprocessed_data_path}\n\n",
    )

    return PostprocessOutputs(
        pyrovelocity_data=FlyteFile(path=str(pyrovelocity_data_path)),
        postprocessed_data=FlyteFile(path=str(postprocessed_data_path)),
        metrics_path=FlyteFile(path=str(metrics_path)),
    )


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=SUMMARIZE_CACHE_VERSION,
    retries=3,
    interruptible=False,
    timeout=timedelta(minutes=120),
    requests=Resources(cpu="8", mem="30Gi", ephemeral_storage="50Gi"),
    limits=Resources(cpu="16", mem="60Gi", ephemeral_storage="200Gi"),
    enable_deck=False,
)
def summarize_data(
    preprocess_data_args: PreprocessDataInterface,
    postprocessing_outputs: PostprocessOutputs,
    training_outputs: TrainingOutputs,
) -> SummarizeOutputs:
    model_path = training_outputs.model_path.download()
    pyrovelocity_data_path = postprocessing_outputs.pyrovelocity_data.download()
    postprocessed_data_path = (
        postprocessing_outputs.postprocessed_data.download()
    )
    run_info_path = training_outputs.run_info_path.download()
    metrics_path = postprocessing_outputs.metrics_path.download()
    loss_plot_path = training_outputs.loss_plot_path.download()
    loss_csv_path = training_outputs.loss_csv_path.download()

    print(
        f"\nmodel_path: {model_path}\n\n",
        f"pyrovelocity_data_path: {pyrovelocity_data_path}\n",
        f"postprocessed_data_path: {postprocessed_data_path}\n",
    )

    data_model_reports_path, dataframe_path = summarize_dataset(
        data_model=training_outputs.data_model,
        data_model_path=training_outputs.data_model_path,
        model_path=model_path,
        pyrovelocity_data_path=pyrovelocity_data_path,
        postprocessed_data_path=postprocessed_data_path,
        cell_state=preprocess_data_args.cell_state,
        vector_field_basis=preprocess_data_args.vector_field_basis,
    )
    print(
        f"\ndata_model_reports_path: {data_model_reports_path}\n",
        f"\ndataframe_path: {dataframe_path}\n\n",
    )

    data_model_metrics_path = Path(data_model_reports_path) / "metrics"
    copy_files_to_reports = [
        run_info_path,
        metrics_path,
        loss_plot_path,
        loss_csv_path,
    ]

    copy_files_result = copy_files_to_directory(
        files_to_copy=copy_files_to_reports,
        target_directory=data_model_metrics_path,
    )
    if isinstance(copy_files_result, Failure):
        print(
            f"\nError copying files to {data_model_reports_path}: {copy_files_result.failure()}\n\n"
        )

    add_duration_to_run_info(run_info_path)
    combined_metrics_path = Path(data_model_reports_path) / "metrics.json"
    combine_json_result = combine_json_files(
        file1=run_info_path,
        file2=metrics_path,
        output_file=combined_metrics_path,
    )
    if isinstance(combine_json_result, Failure):
        print(
            f"\nError combining metrics and run info files in {combined_metrics_path}:\n",
            f"{combine_json_result.failure()}\n\n",
        )

    return SummarizeOutputs(
        data_model_reports=FlyteDirectory(path=str(data_model_reports_path)),
        dataframe=FlyteFile(path=str(dataframe_path)),
        run_metrics_path=FlyteFile(path=str(metrics_path)),
        run_info_path=FlyteFile(path=str(run_info_path)),
        loss_plot_path=FlyteFile(path=str(loss_plot_path)),
        loss_csv_path=FlyteFile(path=str(loss_csv_path)),
        combined_metrics_path=FlyteFile(path=str(combined_metrics_path)),
    )


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=UPLOAD_CACHE_VERSION,
    retries=3,
    interruptible=True,
    timeout=timedelta(minutes=20),
    requests=Resources(cpu="2", mem="4Gi", ephemeral_storage="16Gi"),
    limits=Resources(cpu="8", mem="16Gi", ephemeral_storage="200Gi"),
)
def upload_summary(
    summarize_outputs: SummarizeOutputs,
    training_outputs: TrainingOutputs,
) -> FlyteFile:
    data_model_reports = summarize_outputs.data_model_reports
    reports_path = data_model_reports.download()
    data_model = training_outputs.data_model

    ctx = current_context()
    execution_id = ctx.execution_id.name
    archive_name = f"archive_{data_model}_{execution_id}.tar.gz"

    print(f"\nCreating tarball\n{archive_name}\nfrom {reports_path}...\n\n")

    create_tarball_from_filtered_dir(
        src_dir=reports_path,
        output_filename=archive_name,
        extensions=(
            ".csv",
            ".json",
            ".pdf",
            ".png",
        ),
    )

    upload_result = upload_file_concurrently(
        bucket_name=f"pyrovelocity/reports/{execution_id}",
        source_filename=archive_name,
        destination_blob_name=archive_name,
    )
    if isinstance(upload_result, Success):
        file_url = upload_result.unwrap()
        print(f"\nUpload successful.\nFile URL: {file_url}\n\n")
        return FlyteFile(path=archive_name)
    elif isinstance(upload_result, Failure):
        error = upload_result.failure()
        print(
            f"\nUpload of {archive_name} failed with exception: {error}\n"
            f"Returning empty file.\n\n"
        )
        return FlyteFile(path="")


@dynamic
def map_model_configurations_over_data_set(
    download_dataset_args: DownloadDatasetInterface = DownloadDatasetInterface(),
    preprocess_data_args: PreprocessDataInterface = PreprocessDataInterface(),
    train_model_configuration_1: PyroVelocityTrainInterface = PyroVelocityTrainInterface(),
    train_model_configuration_2: PyroVelocityTrainInterface = PyroVelocityTrainInterface(),
    postprocess_configuration: PostprocessConfiguration = PostprocessConfiguration(),
    train_model_resource_requests: ResourcesJSON = default_training_resource_requests,
    train_model_resource_limits: ResourcesJSON = default_training_resource_limits,
    postprocessing_resource_requests: ResourcesJSON = default_resource_requests,
    postprocessing_resource_limits: ResourcesJSON = default_resource_limits,
    summarizing_resource_requests: ResourcesJSON = default_resource_requests,
    summarizing_resource_limits: ResourcesJSON = default_resource_limits,
    upload_results: bool = True,
) -> list[SummarizeOutputs]:
    """
    Apply the primary workflow to a single dataset with multiple model
    configurations.

    Args:
        download_dataset_args (DownloadDatasetInterface, optional): Configuration for
            pyrovelocity.tasks.data.download_dataset. Defaults to DownloadDatasetInterface().
        preprocess_data_args (PreprocessDataInterface, optional): Configuration for
            pyrovelocity.tasks.preprocess.preprocess_dataset. Defaults to PreprocessDataInterface().
        train_model_configuration_1 (PyroVelocityTrainInterface, optional): Configuration
            for pyrovelocity.train.train_dataset. Defaults to PyroVelocityTrainInterface().
        train_model_configuration_2 (PyroVelocityTrainInterface, optional): Configuration
            for pyrovelocity.train.train_dataset. Defaults to PyroVelocityTrainInterface().
        postprocess_configuration (PostprocessConfiguration, optional): Configuration for
            pyrovelocity.postprocess.postprocess_dataset. Defaults to PostprocessConfiguration().
        train_model_resource_requests (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_training_resource_requests.
        train_model_resource_limits (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_training_resource_limits.
        postprocessing_resource_requests (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_resource_requests.
        postprocessing_resource_limits (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_resource_limits.
        summarizing_resource_requests (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_resource_requests.
        summarizing_resource_limits (ResourcesJSON, optional): Configuration for
            flytekit.Resources. Defaults to default_resource_limits.

    Returns:
        list[FlyteFile]: Workflow outputs as flytekit.types.file.FlyteFile objects.
    """
    data = download_data(download_dataset_args=download_dataset_args)
    processed_outputs = preprocess_data(
        data=data, preprocess_data_args=preprocess_data_args
    )

    train_model_configurations = [
        train_model_configuration_1,
        train_model_configuration_2,
    ]

    model_outputs: list[TrainingOutputs] = []
    dataset_summaries: list[SummarizeOutputs] = []
    for train_model_configuration in train_model_configurations:
        model_output = train_model(
            preprocess_outputs=processed_outputs,
            train_model_configuration=train_model_configuration,
        ).with_overrides(
            requests=Resources(**asdict(train_model_resource_requests)),
            limits=Resources(**asdict(train_model_resource_limits)),
        )
        model_outputs.append(model_output)

        postprocessing_outputs = postprocess_data(
            preprocess_data_args=preprocess_data_args,
            training_outputs=model_output,
            postprocess_configuration=postprocess_configuration,
        ).with_overrides(
            requests=Resources(**asdict(postprocessing_resource_requests)),
            limits=Resources(**asdict(postprocessing_resource_limits)),
        )

        dataset_summary = summarize_data(
            preprocess_data_args=preprocess_data_args,
            postprocessing_outputs=postprocessing_outputs,
            training_outputs=model_output,
        ).with_overrides(
            requests=Resources(**asdict(summarizing_resource_requests)),
            limits=Resources(**asdict(summarizing_resource_limits)),
        )

        if upload_results:
            upload_summary(
                summarize_outputs=dataset_summary,
                training_outputs=model_output,
            )

        dataset_summaries.append(dataset_summary)

    return dataset_summaries


@task(
    cache=PYROVELOCITY_CACHE_FLAG,
    cache_version=COMBINE_METRICS_CACHE_VERSION,
    retries=3,
    interruptible=True,
    timeout=timedelta(minutes=20),
    requests=Resources(cpu="2", mem="4Gi", ephemeral_storage="8Gi"),
    limits=Resources(cpu="4", mem="8Gi", ephemeral_storage="16Gi"),
)
def combine_all_metrics(
    results: List[List[SummarizeOutputs]]
) -> CombinedMetricsOutputs:
    combined_metrics = {}

    for dataset_results in results:
        for model_result in dataset_results:
            metrics_path = model_result.combined_metrics_path.download()
            metrics_result = load_json(Path(metrics_path))

            if isinstance(metrics_result, Success):
                metrics = metrics_result.unwrap()
                run_name = metrics.get("run_name", "unknown")
                combined_metrics[run_name] = metrics
            else:
                print(
                    f"Failed to load metrics from {metrics_path}: {metrics_result.failure()}"
                )

    json_metrics_file = Path("combined_metrics.json")
    with json_metrics_file.open("w") as f:
        json.dump(combined_metrics, f, indent=2)

    latex_table, html_table, markdown_table, _ = generate_tables(
        combined_metrics
    )

    latex_metrics_file = Path("combined_metrics_table.tex")
    with latex_metrics_file.open("w") as f:
        f.write(latex_table)

    html_metrics_file = Path("combined_metrics_table.html")
    with html_metrics_file.open("w") as f:
        f.write(html_table)

    md_metrics_file = Path("combined_metrics_table.md")
    with md_metrics_file.open("w") as f:
        f.write(markdown_table)

    ctx = current_context()
    execution_id = ctx.execution_id.name

    files_to_upload = [
        json_metrics_file,
        latex_metrics_file,
        html_metrics_file,
        md_metrics_file,
    ]
    upload_results = []

    for file in files_to_upload:
        upload_result = upload_file_concurrently(
            bucket_name=f"pyrovelocity/reports/{execution_id}",
            source_filename=file,
            destination_blob_name=str(file),
        )
        upload_results.append(upload_result)

    if all(isinstance(result, Success) for result in upload_results):
        print("\nAll uploads successful.")
        return CombinedMetricsOutputs(
            json_metrics=FlyteFile(path=str(json_metrics_file)),
            latex_metrics=FlyteFile(path=str(latex_metrics_file)),
            html_metrics=FlyteFile(path=str(html_metrics_file)),
            md_metrics=FlyteFile(path=str(md_metrics_file)),
        )
    else:
        print("\nOne or more uploads failed.")
        failed_uploads = [
            str(file)
            for file, result in zip(files_to_upload, upload_results)
            if isinstance(result, Failure)
        ]
        print(f"Failed uploads: {', '.join(failed_uploads)}")
        return CombinedMetricsOutputs(
            json_metrics=FlyteFile(path=""),
            latex_metrics=FlyteFile(path=""),
            html_metrics=FlyteFile(path=""),
            md_metrics=FlyteFile(path=""),
        )


@dynamic
def training_workflow(
    simulated_configuration: WorkflowConfiguration = simulated_configuration,
    pancreas_configuration: WorkflowConfiguration = pancreas_configuration,
    bonemarrow_configuration: WorkflowConfiguration = bonemarrow_configuration,
    pbmc5k_configuration: WorkflowConfiguration = pbmc5k_configuration,
    pbmc10k_configuration: WorkflowConfiguration = pbmc10k_configuration,
    pbmc68k_configuration: WorkflowConfiguration = pbmc68k_configuration,
    pons_configuration: WorkflowConfiguration = pons_configuration,
    larry_configuration: WorkflowConfiguration = larry_configuration,
    larry_neu_configuration: WorkflowConfiguration = larry_neu_configuration,
    larry_mono_configuration: WorkflowConfiguration = larry_mono_configuration,
    larry_multilineage_configuration: WorkflowConfiguration = larry_multilineage_configuration,
) -> list[list[SummarizeOutputs]]:
    """
    Apply the primary workflow to a collection of configurations.
    Conditionally executes configurations based on the value of PYROVELOCITY_DATA_SUBSET.
    """
    results = []
    configurations = [
        (simulated_configuration, "simulated"),
    ]

    if not PYROVELOCITY_DATA_SUBSET:
        configurations += [
            (pancreas_configuration, "pancreas"),
            (bonemarrow_configuration, "bonemarrow"),
            (pbmc5k_configuration, "pbmc5k"),
            (pbmc10k_configuration, "pbmc10k"),
            # (pbmc68k_configuration, "pbmc68k"),
            (pons_configuration, "pons"),
            # (larry_configuration, "larry"),
            (larry_neu_configuration, "larry_neu"),
            (larry_mono_configuration, "larry_mono"),
            (larry_multilineage_configuration, "larry_multilineage"),
        ]

    for config, _ in configurations:
        result = map_model_configurations_over_data_set(
            download_dataset_args=config.download_dataset,
            preprocess_data_args=config.preprocess_data,
            train_model_configuration_1=config.training_configuration_1,
            train_model_configuration_2=config.training_configuration_2,
            postprocess_configuration=config.postprocess_configuration,
            train_model_resource_requests=config.training_resources_requests,
            train_model_resource_limits=config.training_resources_limits,
            postprocessing_resource_requests=config.postprocessing_resources_requests,
            postprocessing_resource_limits=config.postprocessing_resources_limits,
            summarizing_resource_requests=config.summarizing_resources_requests,
            summarizing_resource_limits=config.summarizing_resources_limits,
            upload_results=config.upload_results,
        )
        results.append(result)

    combine_all_metrics(results=results)

    return results


if __name__ == "__main__":
    """
    The module workflow is applied to a single dataset together with a list of
    model configurations.

    There are three ways to execute the train_model() task, which impacts how
    subsequent tasks are executed. This can also be executed with the @workflow
    decorator for a single model configuration

    ```python
    train_model_configuration: PyroVelocityTrainInterface = PyroVelocityTrainInterface(),
    ) -> TrainingOutputs:
    ...
    model_outputs = train_model(
        data=processed_data,
        train_model_configuration=train_model_configuration,
    )
    ```

    or for multiple model configurations using a map task

    ```python
    import functools
    from flytekit.experimental import map_task

    partial_train_model = functools.partial(train_model, data=processed_data)
    model_outputs = map_task(partial_train_model)(
        train_model_args=train_model_configurations,
    )
    ```

    The dynamic workflow is preferred to support both mapping of tasks over
    multiple model configurations and overriding resources from configuration
    data.
    """
    print(f"Running training_workflow() { training_workflow() }")



import pytest

from pyrovelocity.plots import plot_gene_selection_summary


@pytest.mark.slow
@pytest.mark.integration
def test_model2_plot_gene_selection_summary(
    postprocessed_model2_data,
    posterior_samples_model2,
    putative_model2_marker_genes,
    data_model2_reports_path,
):
    plot_gene_selection_summary(
        adata=postprocessed_model2_data,
        posterior_samples=posterior_samples_model2,
        basis="umap",
        cell_state="leiden",
        plot_name=data_model2_reports_path / "gene_selection_summary_plot.pdf",
        selected_genes=putative_model2_marker_genes,
        show_marginal_histograms=False,
    )


@pytest.mark.slow
@pytest.mark.integration
def test_model1_plot_gene_selection_summary(
    postprocessed_model1_data,
    posterior_samples_model1,
    putative_model1_marker_genes,
    data_model1_reports_path,
):
    plot_gene_selection_summary(
        adata=postprocessed_model1_data,
        posterior_samples=posterior_samples_model1,
        basis="umap",
        cell_state="leiden",
        plot_name=data_model1_reports_path / "gene_selection_summary_plot.pdf",
        selected_genes=putative_model1_marker_genes,
        show_marginal_histograms=False,
    )



"""Tests for `pyrovelocity.tasks.summarize` module."""


import pytest


def test_load_summarize():
    from pyrovelocity.tasks import summarize

    print(summarize.__file__)


@pytest.mark.slow
@pytest.mark.integration
def test_summarize_dataset(summarize_dataset_output):
    return summarize_dataset_output


@pytest.mark.slow
@pytest.mark.integration
def test_summarize_dataset_model1(summarize_dataset_model1_output):
    return summarize_dataset_model1_output



from typing import List

import pandas as pd
import pytest

from pyrovelocity.analysis.analyze import (
    pareto_frontier_genes,
    top_mae_genes,
)


@pytest.fixture
def sample_volcano_data() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "mean_mae": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8],
            "time_correlation": [-0.9, -0.6, -0.3, 0, 0.3, 0.6, 0.9, 1.0],
        },
        index=[f"Gene{i}" for i in range(1, 9)],
    )


@pytest.fixture
def sample_volcano_data_with_ribosomal() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "mean_mae": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            "time_correlation": [
                -0.9,
                -0.6,
                -0.3,
                0,
                0.3,
                0.6,
                0.9,
                1.0,
                -0.5,
                0.5,
            ],
        },
        index=[
            "Gene1",
            "Rpl1",
            "Gene3",
            "Rps2",
            "Gene5",
            "Gene6",
            "Gene7",
            "Gene8",
            "Gene9",
            "Gene10",
        ],
    )


def test_pareto_frontier_genes_basic(sample_volcano_data):
    result = pareto_frontier_genes(
        sample_volcano_data, num_genes=3, max_iters=1000
    )
    assert isinstance(result, List)
    assert len(result) == 3
    assert all(gene in sample_volcano_data.index for gene in result)


def test_pareto_frontier_genes_order(sample_volcano_data):
    result = pareto_frontier_genes(
        sample_volcano_data, num_genes=8, max_iters=1000
    )
    time_correlations = [
        sample_volcano_data.loc[gene, "time_correlation"] for gene in result
    ]
    assert time_correlations == sorted(time_correlations, reverse=True)
    assert len(result) == 8


def test_pareto_frontier_genes_ribosomal_filtering(
    sample_volcano_data_with_ribosomal
):
    result = pareto_frontier_genes(
        sample_volcano_data_with_ribosomal, num_genes=6, max_iters=1000
    )
    assert "Rpl1" not in result
    assert "Rps2" not in result
    assert len(result) == 6


def test_pareto_frontier_genes_max_iters(sample_volcano_data):
    result = pareto_frontier_genes(
        sample_volcano_data, num_genes=8, max_iters=1
    )
    assert len(result) < 8


def test_pareto_frontier_genes_fewer_genes(sample_volcano_data):
    result = pareto_frontier_genes(
        sample_volcano_data, num_genes=10, max_iters=1000
    )
    assert len(result) == 8


def test_top_mae_genes_basic(sample_volcano_data):
    result = top_mae_genes(
        sample_volcano_data, mae_top_percentile=50, min_genes_per_bin=1
    )
    assert isinstance(result, List)
    assert 4 <= len(result) <= 8
    assert all(gene in sample_volcano_data.index for gene in result)


def test_top_mae_genes_order(sample_volcano_data):
    result = top_mae_genes(
        sample_volcano_data, mae_top_percentile=100, min_genes_per_bin=1
    )
    assert result == [
        "Gene8",
        "Gene7",
        "Gene6",
        "Gene5",
        "Gene4",
        "Gene3",
        "Gene2",
        "Gene1",
    ]


def test_top_mae_genes_ribosomal_filtering(sample_volcano_data_with_ribosomal):
    result = top_mae_genes(
        sample_volcano_data_with_ribosomal,
        mae_top_percentile=50,
        min_genes_per_bin=1,
    )
    assert "Rpl1" not in result
    assert "Rps2" not in result
    assert 4 <= len(result) <= 8


def test_top_mae_genes_min_genes_per_bin(sample_volcano_data):
    result = top_mae_genes(
        sample_volcano_data,
        mae_top_percentile=10,
        min_genes_per_bin=2,
    )
    assert len(result) >= 8


def test_top_mae_genes_time_correlation_sorting(sample_volcano_data):
    result = top_mae_genes(
        sample_volcano_data, mae_top_percentile=50, min_genes_per_bin=1
    )
    time_correlations = [
        sample_volcano_data.loc[gene, "time_correlation"] for gene in result
    ]
    assert time_correlations == sorted(time_correlations, reverse=True)


def test_top_mae_genes_percentile_threshold(sample_volcano_data):
    result_10 = top_mae_genes(
        sample_volcano_data, mae_top_percentile=10, min_genes_per_bin=1
    )
    result_50 = top_mae_genes(
        sample_volcano_data, mae_top_percentile=50, min_genes_per_bin=1
    )
    assert len(result_10) <= len(result_50)


def test_top_mae_genes_invalid_percentile():
    with pytest.raises(ValueError):
        top_mae_genes(
            pd.DataFrame(), mae_top_percentile=101, min_genes_per_bin=1
        )
    with pytest.raises(ValueError):
        top_mae_genes(pd.DataFrame(), mae_top_percentile=0, min_genes_per_bin=1)


def test_top_mae_genes_invalid_min_genes():
    with pytest.raises(ValueError):
        top_mae_genes(
            pd.DataFrame(), mae_top_percentile=50, min_genes_per_bin=-1
        )



import pytest

from pyrovelocity.plots import plot_vector_field_summary


@pytest.mark.slow
@pytest.mark.integration
def test_model2_plot_vector_field_summary(
    postprocessed_model2_data,
    posterior_samples_model2,
    data_model2_reports_path,
):
    vector_field_summary_plot = (
        data_model2_reports_path / "vector_field_summary_plot.pdf"
    )
    plot_vector_field_summary(
        adata=postprocessed_model2_data,
        posterior_samples=posterior_samples_model2,
        vector_field_basis="umap",
        plot_name=vector_field_summary_plot,
        cell_state="leiden",
    )


@pytest.mark.slow
@pytest.mark.integration
def test_model1_plot_vector_field_summary(
    postprocessed_model1_data,
    posterior_samples_model1,
    data_model1_reports_path,
):
    vector_field_summary_plot = (
        data_model1_reports_path / "vector_field_summary_plot.pdf"
    )
    plot_vector_field_summary(
        adata=postprocessed_model1_data,
        posterior_samples=posterior_samples_model1,
        vector_field_basis="umap",
        plot_name=vector_field_summary_plot,
        cell_state="leiden",
    )



import tempfile

import numpy as np
import pandas as pd
import pytest
from anndata import AnnData
from scipy import sparse

from pyrovelocity.io.serialization import (
    create_sample_anndata,
    deserialize_anndata,
    load_anndata_from_json,
    save_anndata_to_json,
    serialize_anndata,
)


@pytest.fixture
def sample_adata():
    return create_sample_anndata(5, 4)


def test_create_sample_anndata():
    adata = create_sample_anndata(5, 4)
    assert isinstance(adata, AnnData)
    assert adata.n_obs == 5
    assert adata.n_vars == 4
    assert "batch" in adata.obs.columns
    assert "a" in adata.obsm
    assert "a" in adata.layers


def test_serialize_deserialize_anndata(sample_adata):
    serialized = serialize_anndata(sample_adata)
    deserialized = deserialize_anndata(serialized)

    assert np.allclose(sample_adata.X, deserialized.X)
    assert sample_adata.obs.equals(deserialized.obs)
    assert sample_adata.var.equals(deserialized.var)
    assert np.allclose(sample_adata.obsm["a"], deserialized.obsm["a"])
    assert np.allclose(sample_adata.layers["a"], deserialized.layers["a"])
    assert sample_adata.uns == deserialized.uns


def test_save_load_anndata_json(sample_adata, tmp_path):
    file_path = tmp_path / "test_adata.json"
    save_anndata_to_json(sample_adata, file_path)
    assert file_path.exists()

    loaded_adata = load_anndata_from_json(file_path)
    assert np.allclose(sample_adata.X, loaded_adata.X)
    assert sample_adata.obs.equals(loaded_adata.obs)
    assert sample_adata.var.equals(loaded_adata.var)
    assert np.allclose(sample_adata.obsm["a"], loaded_adata.obsm["a"])
    assert np.allclose(sample_adata.layers["a"], loaded_adata.layers["a"])
    assert sample_adata.uns == loaded_adata.uns


def test_serialize_large_anndata():
    large_adata = create_sample_anndata(1000, 100)
    serialized = serialize_anndata(large_adata)
    assert isinstance(serialized, dict)
    assert all(
        key in serialized
        for key in ["X", "obs", "var", "obsm", "layers", "uns"]
    )


def test_deserialize_invalid_data():
    invalid_data = {
        "X": [[1, 2], [3, 4]],
        # "obs": {},
        "var": {},
        "uns": {},
        "obsm": {},
        "varm": {},
        "layers": {},
        # "shape": (2, 2),
    }
    with pytest.raises(ValueError):
        deserialize_anndata(invalid_data)


def test_save_load_empty_anndata(tmp_path):
    empty_adata = AnnData(
        X=np.empty((0, 0)),
        obs=pd.DataFrame(index=[]),
        var=pd.DataFrame(index=[]),
    )
    file_path = tmp_path / "empty_adata.json"
    save_anndata_to_json(empty_adata, file_path)
    loaded_adata = load_anndata_from_json(file_path)
    assert loaded_adata.n_obs == 0
    assert loaded_adata.n_vars == 0


def test_serialize_deserialize_sparse_matrix():
    sparse_adata = AnnData(X=sparse.random(100, 100, density=0.1, format="csr"))
    sparse_adata.X[sparse_adata.X < 0.9] = 0
    sparse_adata.X = sparse_adata.X.tocsr()

    serialized = serialize_anndata(sparse_adata)
    deserialized = deserialize_anndata(serialized)

    assert np.allclose(sparse_adata.X.toarray(), deserialized.X)


def test_serialize_deserialize_with_complex_uns():
    adata = create_sample_anndata(100, 50)
    adata.uns["complex_item"] = {
        "nested": {"a": 1, "b": [1, 2, 3]},
        "array": np.array([1, 2, 3]),
    }

    serialized = serialize_anndata(adata)
    deserialized = deserialize_anndata(serialized)

    assert (
        deserialized.uns["complex_item"]["nested"]
        == adata.uns["complex_item"]["nested"]
    )
    assert np.array_equal(
        deserialized.uns["complex_item"]["array"],
        adata.uns["complex_item"]["array"],
    )


def test_load_anndata_from_nonexistent_file():
    with pytest.raises(FileNotFoundError):
        load_anndata_from_json("nonexistent_file.json")


def test_save_anndata_to_invalid_path(sample_adata):
    with pytest.raises(OSError):
        save_anndata_to_json(sample_adata, "/invalid/path/adata.json")


def test_roundtrip_with_all_attributes():
    adata = create_sample_anndata(10, 5)
    adata.raw = adata.copy()
    adata.obsp["distances"] = np.random.random((10, 10))
    adata.varp["gene_correlation"] = np.random.random((5, 5))

    with tempfile.NamedTemporaryFile(
        mode="w+", suffix=".json", delete=False
    ) as tmp:
        save_anndata_to_json(adata, tmp.name)
        loaded_adata = load_anndata_from_json(tmp.name)

    assert np.allclose(adata.X, loaded_adata.X)
    assert adata.obs.equals(loaded_adata.obs)
    assert adata.var.equals(loaded_adata.var)
    assert np.allclose(adata.raw.X, loaded_adata.raw.X)
    assert np.allclose(adata.obsp["distances"], loaded_adata.obsp["distances"])
    assert np.allclose(
        adata.varp["gene_correlation"], loaded_adata.varp["gene_correlation"]
    )



import numpy as np
from beartype import beartype
from beartype.typing import Any, Dict, Tuple
from sparse import COO

__all__ = [
    "analyze_sparsification",
    "calculate_density",
    "check_sparsification_candidates",
    "densify_arrays",
    "print_sparsification_results",
    "sparsify_arrays",
]


@beartype
def calculate_density(arr: np.ndarray | COO) -> float:
    if isinstance(arr, COO):
        return arr.density

    total_elements = arr.size
    if total_elements == 0:
        return 0.0

    non_zero_count = np.count_nonzero(arr)
    return non_zero_count / total_elements


@beartype
def check_sparsification_candidates(
    data_dict: Dict[str, np.ndarray | COO],
    density_threshold: float = 0.5,
) -> Dict[str, Dict[str, Any]]:
    candidates = {}
    for key, arr in data_dict.items():
        density = calculate_density(arr)
        if density <= density_threshold:
            sparse_arr = arr if isinstance(arr, COO) else COO.from_numpy(arr)
            original_size = (
                arr.nbytes if isinstance(arr, np.ndarray) else sparse_arr.nbytes
            )
            candidates[key] = {
                "shape": arr.shape,
                "density": density,
                "original_size": original_size,
                "sparse_size": sparse_arr.nbytes,
                "size_reduction": 1 - (sparse_arr.nbytes / original_size)
                if original_size > 0
                else 0,
            }
    return candidates


@beartype
def print_sparsification_results(candidates: Dict[str, Dict[str, Any]]) -> None:
    for key, info in candidates.items():
        print(f"Array '{key}':")
        print(f"  Shape: {info['shape']}")
        print(f"  Density: {info['density']:.2%}")
        print(f"  Original size: {info['original_size']} bytes")
        print(f"  Sparse size: {info['sparse_size']} bytes")
        print(f"  Size reduction: {info['size_reduction']} bytes")
        print()


@beartype
def analyze_sparsification(
    data_dict: Dict[str, np.ndarray],
    density_threshold: float = 0.5,
) -> Dict[str, Dict[str, Any]]:
    candidates = check_sparsification_candidates(data_dict, density_threshold)
    print_sparsification_results(candidates)
    return candidates


@beartype
def sparsify_arrays(
    data_dict: Dict[str, np.ndarray | COO],
    density_threshold: float = 0.6,
) -> Dict[str, np.ndarray | COO]:
    candidates = analyze_sparsification(data_dict, density_threshold)
    result = {}
    for key, arr in data_dict.items():
        if key in candidates and not isinstance(arr, COO):
            result[key] = COO.from_numpy(arr)
        else:
            result[key] = arr
    return result


@beartype
def densify_arrays(
    data_dict: Dict[str, np.ndarray | COO]
) -> Dict[str, np.ndarray]:
    result = {}
    for key, arr in data_dict.items():
        if isinstance(arr, COO):
            result[key] = arr.todense()
        else:
            result[key] = arr
    return result



import os

import numpy as np
import pandas as pd
import pytest
from sparse import COO

from pyrovelocity.io.compressedpickle import CompressedPickle


@pytest.fixture(scope="module")
def test_data():
    return pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})


@pytest.fixture(scope="module")
def test_file():
    return "test_data.pkl.zst"


def test_save(test_data, test_file):
    CompressedPickle.save(test_file, test_data)
    assert os.path.exists(test_file)


def test_load(test_data, test_file):
    loaded_data = CompressedPickle.load(test_file)
    assert loaded_data.equals(test_data)


def compare_dicts(dict1, dict2):
    if dict1.keys() != dict2.keys():
        return False

    for key in dict1:
        if isinstance(dict1[key], np.ndarray):
            if not np.array_equal(dict1[key], dict2[key]):
                return False
        elif dict1[key] != dict2[key]:
            return False

    return True


def test_save_load_dict(test_file):
    test_dict = {"a": np.arange(5), "b": np.linspace(0, 1, 5)}
    CompressedPickle.save(test_file, test_dict)
    loaded_dict = CompressedPickle.load(test_file)
    assert compare_dicts(test_dict, loaded_dict)


def compare_complex_objects(obj1, obj2):
    if len(obj1) != len(obj2):
        return False

    for item1, item2 in zip(obj1, obj2):
        if isinstance(item1, dict) and isinstance(item2, dict):
            if not compare_dicts(item1, item2):
                return False
        elif isinstance(item1, np.ndarray) and isinstance(item2, np.ndarray):
            if not np.array_equal(item1, item2):
                return False
        elif item1 != item2:
            return False

    return True


def test_save_load_complex_object(test_file):
    test_obj = [{"a": np.array([1, 2, 3]), "b": 1.5}, (4, "test", [1, 2])]
    CompressedPickle.save(test_file, test_obj)
    loaded_obj = CompressedPickle.load(test_file)
    assert compare_complex_objects(test_obj, loaded_obj)


def test_load_non_existent_file():
    with pytest.raises(FileNotFoundError):
        CompressedPickle.load("non_existent.pkl.zst")


@pytest.fixture(scope="module", autouse=True)
def cleanup(request, test_file):
    def delete_test_file():
        if os.path.exists(test_file):
            os.remove(test_file)

    request.addfinalizer(delete_test_file)


@pytest.fixture
def sparse_dict():
    return {
        "dense": np.array([[1, 2], [3, 4]]),
        "sparse": np.array([[1, 0], [0, 2]]),
        "very_sparse": np.array([[0, 1], [0, 0]]),
    }


def test_save_load_sparse_dict(test_file, sparse_dict):
    CompressedPickle.save(test_file, sparse_dict)
    loaded_dict = CompressedPickle.load(test_file)

    assert isinstance(loaded_dict["dense"], np.ndarray)
    assert isinstance(loaded_dict["sparse"], np.ndarray)
    assert isinstance(loaded_dict["very_sparse"], np.ndarray)

    assert np.array_equal(loaded_dict["dense"], sparse_dict["dense"])
    assert np.array_equal(loaded_dict["sparse"], sparse_dict["sparse"])
    assert np.array_equal(
        loaded_dict["very_sparse"], sparse_dict["very_sparse"]
    )


def test_save_load_sparse_dict_without_densify(test_file, sparse_dict):
    CompressedPickle.save(test_file, sparse_dict)
    loaded_dict = CompressedPickle.load(test_file, densify=False)

    assert isinstance(loaded_dict["dense"], np.ndarray)
    assert isinstance(loaded_dict["sparse"], COO)
    assert isinstance(loaded_dict["very_sparse"], COO)

    assert np.array_equal(loaded_dict["dense"], sparse_dict["dense"])
    assert np.array_equal(
        loaded_dict["sparse"].todense(), sparse_dict["sparse"]
    )
    assert np.array_equal(
        loaded_dict["very_sparse"].todense(), sparse_dict["very_sparse"]
    )


def test_save_without_sparsify(test_file, sparse_dict):
    CompressedPickle.save(test_file, sparse_dict, sparsify=False)
    loaded_dict = CompressedPickle.load(test_file)

    assert all(isinstance(arr, np.ndarray) for arr in loaded_dict.values())
    assert all(
        np.array_equal(loaded_dict[key], sparse_dict[key])
        for key in sparse_dict
    )


# TODO: Handle sparsification when values are not exclusively arrays
def test_save_load_mixed_dict(test_file):
    mixed_dict = {
        "dense": np.array([[1, 2], [3, 4]]),
        "sparse": COO.from_numpy(np.array([[1, 0], [0, 2]])),
        "very_sparse": COO.from_numpy(np.array([[0, 1], [0, 0]])),
        "non_array": "test string",
    }

    CompressedPickle.save(test_file, mixed_dict)
    loaded_dict = CompressedPickle.load(test_file)

    assert isinstance(loaded_dict["dense"], np.ndarray)
    # assert isinstance(loaded_dict["sparse"], np.ndarray)
    # assert isinstance(loaded_dict["very_sparse"], np.ndarray)
    assert isinstance(loaded_dict["sparse"], COO)
    assert isinstance(loaded_dict["very_sparse"], COO)
    assert isinstance(loaded_dict["non_array"], str)

    assert np.array_equal(loaded_dict["dense"], mixed_dict["dense"])
    # assert np.array_equal(loaded_dict["sparse"], mixed_dict["sparse"].todense())
    # assert np.array_equal(
    #     loaded_dict["very_sparse"], mixed_dict["very_sparse"].todense()
    # )
    assert np.array_equal(
        loaded_dict["sparse"].todense(),
        mixed_dict["sparse"].todense(),
    )
    assert np.array_equal(
        loaded_dict["very_sparse"].todense(),
        mixed_dict["very_sparse"].todense(),
    )
    assert loaded_dict["non_array"] == mixed_dict["non_array"]


def test_save_load_non_dict(test_file):
    non_dict_obj = np.array([[1, 2], [3, 4]])
    CompressedPickle.save(test_file, non_dict_obj)
    loaded_obj = CompressedPickle.load(test_file)

    assert isinstance(loaded_obj, np.ndarray)
    assert np.array_equal(loaded_obj, non_dict_obj)



import pytest

from pyrovelocity.plots import rainbowplot


@pytest.mark.slow
@pytest.mark.integration
def test_model2_rainbowplot(
    postprocessed_model2_data,
    posterior_samples_model2,
    putative_model2_marker_genes,
    data_model2_reports_path,
):
    rainbowplot(
        volcano_data=posterior_samples_model2["gene_ranking"],
        adata=postprocessed_model2_data,
        posterior_samples=posterior_samples_model2,
        genes=putative_model2_marker_genes,
        data=["st", "ut"],
        basis="umap",
        cell_state="leiden",
        save_plot=True,
        rainbow_plot_path=data_model2_reports_path
        / "gene_selection_rainbow_plot.pdf",
    )


@pytest.mark.slow
@pytest.mark.integration
def test_model1_rainbowplot(
    postprocessed_model1_data,
    posterior_samples_model1,
    putative_model1_marker_genes,
    data_model1_reports_path,
):
    rainbowplot(
        volcano_data=posterior_samples_model1["gene_ranking"],
        adata=postprocessed_model1_data,
        posterior_samples=posterior_samples_model1,
        genes=putative_model1_marker_genes,
        data=["st", "ut"],
        basis="umap",
        cell_state="leiden",
        save_plot=True,
        rainbow_plot_path=data_model1_reports_path
        / "gene_selection_rainbow_plot.pdf",
    )



from pyrovelocity.plots import plot_spliced_unspliced_histogram


def test_plot_spliced_unspliced_histogram(adata_preprocessed, tmp_path):
    chart = plot_spliced_unspliced_histogram(adata_preprocessed)
    chart.save(tmp_path / "spliced_unspliced_histogram.pdf")
    chart.save(tmp_path / "spliced_unspliced_histogram.svg")
    chart.save(tmp_path / "spliced_unspliced_histogram.png")
    chart.save(tmp_path / "spliced_unspliced_histogram.html")



from typing import Any, Dict, Optional, Tuple
from unittest.mock import Mock

import pytest
from beartype import beartype
from flytekit.configuration import Config
from omegaconf import DictConfig
from pytest_mock import MockerFixture

from pyrovelocity.io.cluster import get_remote_task_results


class MockPrimitive:
    def __init__(self, value: Any):
        self.string_value: Optional[str] = (
            value if isinstance(value, str) else None
        )
        self.integer: Optional[int] = value if isinstance(value, int) else None
        self.float_value: Optional[float] = (
            value if isinstance(value, float) else None
        )
        self.boolean: Optional[bool] = (
            value if isinstance(value, bool) else None
        )


class MockScalar:
    def __init__(self, value: Any):
        self.primitive: MockPrimitive = MockPrimitive(value)


class MockLiteral:
    def __init__(self, value: Any):
        self.scalar: MockScalar = MockScalar(value)


class MockLiteralMap:
    def __init__(self, literals: Dict[str, Any]):
        self.literals: Dict[str, MockLiteral] = {
            k: MockLiteral(v) for k, v in literals.items()
        }


class MockFlyteRemote:
    @beartype
    def __init__(self, config: Any) -> None:
        self.config: Any = config

    @beartype
    def get(self, uri: str) -> MockLiteralMap:
        if uri.endswith("/i"):
            return MockLiteralMap({"input_key": "input_value"})
        elif uri.endswith("/o"):
            return MockLiteralMap({"output_key": "output_value"})
        raise ValueError(f"Unexpected URI: {uri}")


@pytest.fixture
@beartype
def mock_flyte_remote(mocker: MockerFixture) -> Mock:
    mock_config = mocker.Mock()
    mocker.patch.object(Config, "for_endpoint", return_value=mock_config)
    mock_remote = mocker.Mock(spec=MockFlyteRemote)
    mock_remote.get.side_effect = MockFlyteRemote(mock_config).get
    mocker.patch(
        "pyrovelocity.io.cluster.FlyteRemote", return_value=mock_remote
    )
    return mock_remote


@pytest.fixture
@beartype
def mock_print_config_tree(mocker: MockerFixture) -> Mock:
    return mocker.patch("pyrovelocity.io.cluster.print_config_tree")


@pytest.mark.parametrize(
    "execution_id, task_id, endpoint, protocol, project, domain",
    [
        (
            "exec-1",
            "task-1",
            "custom.endpoint",
            "protocol://v1",
            "custom-project",
            "custom-domain",
        ),
        (
            "exec-2",
            "task-2",
            "flyte.cluster.testorg.net",
            "flyte://v1",
            "testorg",
            "development",
        ),
    ],
)
@beartype
def test_get_remote_task_results(
    mocker: MockerFixture,
    mock_flyte_remote: Mock,
    mock_print_config_tree: Mock,
    execution_id: str,
    task_id: str,
    endpoint: str,
    protocol: str,
    project: str,
    domain: str,
) -> None:
    result: Tuple[DictConfig, DictConfig] = get_remote_task_results(
        execution_id, task_id, endpoint, protocol, project, domain
    )

    assert isinstance(result, Tuple)
    assert len(result) == 2
    assert all(isinstance(item, DictConfig) for item in result)

    inputs, outputs = result

    Config.for_endpoint.assert_called_once_with(endpoint=endpoint)

    expected_base_uri: str = (
        f"{protocol}/{project}/{domain}/{execution_id}/{task_id}"
    )
    mock_flyte_remote.get.assert_any_call(f"{expected_base_uri}/i")
    mock_flyte_remote.get.assert_any_call(f"{expected_base_uri}/o")

    assert "input_key" in inputs
    assert "output_key" in outputs
    assert inputs["input_key"] == "input_value"
    assert outputs["output_key"] == "output_value"

    assert mock_print_config_tree.call_count == 2
    mock_print_config_tree.assert_any_call({"input_key": "input_value"})
    mock_print_config_tree.assert_any_call({"output_key": "output_value"})


@beartype
def test_get_remote_task_results_type_checking() -> None:
    with pytest.raises(Exception):
        get_remote_task_results(123, "task_id")

    with pytest.raises(Exception):
        get_remote_task_results("execution_id", 123)

    with pytest.raises(Exception):
        get_remote_task_results("execution_id", "task_id", endpoint=123)



"""Tests for `pyrovelocity.tasks.preprocess` module."""
import pytest

from pyrovelocity.tasks.preprocess import preprocess_dataset


def test_load_preprocess():
    from pyrovelocity.tasks import preprocess

    print(preprocess.__file__)


@pytest.mark.slow
@pytest.mark.integration
def test_preprocess_dataset(preprocess_dataset_output):
    return preprocess_dataset_output



from pathlib import Path
from typing import Dict, List, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scvelo as scv
import seaborn as sns
from anndata import AnnData
from beartype import beartype
from matplotlib import ticker
from matplotlib.figure import Figure, FigureBase
from numpy import ndarray
from pandas import DataFrame, Index

from pyrovelocity.plots._common import set_colorbar, set_font_size
from pyrovelocity.utils import ensure_numpy_array

__all__ = ["rainbowplot", "us_rainbowplot"]


@beartype
def rainbowplot(
    volcano_data: DataFrame,
    adata: AnnData,
    posterior_samples: Dict[str, ndarray],
    fig: Optional[FigureBase] = None,
    genes: Optional[List[str]] = None,
    data: List[str] = ["st", "ut"],
    cell_state: str = "clusters",
    basis: str = "umap",
    num_genes: int = 5,
    add_line: bool = True,
    negative_correlation: bool = False,
    scvelo_colors: bool = False,
    save_plot: bool = False,
    show_data: bool = True,
    rainbow_plot_path: str | Path = "rainbow.pdf",
) -> FigureBase:
    set_font_size(7)

    if genes is None:
        genes = get_genes(volcano_data, num_genes, negative_correlation)
        genes = genes.tolist()
    number_of_genes = len(genes)

    subplot_height = 1

    if show_data:
        horizontal_panels = 4
    else:
        horizontal_panels = 3

    subplot_width = 2.0 * subplot_height * horizontal_panels

    if fig is None:
        fig, ax = plt.subplots(
            number_of_genes,
            horizontal_panels,
            figsize=(subplot_width, subplot_height * number_of_genes),
        )
    else:
        ax = fig.subplots(number_of_genes, horizontal_panels)

    if scvelo_colors:
        colors = setup_scvelo_colors(adata, cell_state)
    else:
        colors = setup_colors(adata, cell_state)

    st, ut = get_posterior_samples_mean(data, posterior_samples)

    for n, gene in enumerate(genes):
        ress = get_data(gene, st, ut, adata, cell_state, posterior_samples)
        ax1 = ax[n, 1]
        ax2 = ax[n, 0]
        ax3 = ax[n, 2]
        if show_data:
            ax4 = ax[n, 3]

        if n == 0:
            ax1.set_title("Predictive dynamics", fontsize=7)
            ax2.set_title("Phase portrait", fontsize=7)
            ax3.set_title("Predictive spliced", fontsize=7)
            if show_data:
                ax4.set_title("Log spliced data", fontsize=7)

        plot_gene(ax1, ress, colors, add_line)
        scatterplot(ax2, ress, colors)
        (index,) = np.where(adata.var_names == gene)
        im = ax3.scatter(
            adata.obsm[f"X_{basis}"][:, 0],
            adata.obsm[f"X_{basis}"][:, 1],
            s=3,
            c=st[:, index].flatten(),
            cmap="cividis",
        )
        set_colorbar(im, ax3, labelsize=5, fig=fig, rainbow=True)
        ax3.axis("off")
        if show_data:
            im = ax4.scatter(
                adata.obsm[f"X_{basis}"][:, 0],
                adata.obsm[f"X_{basis}"][:, 1],
                s=3,
                c=ensure_numpy_array(adata.X[:, index]).flatten(),
                cmap="cividis",
            )
            set_colorbar(im, ax4, labelsize=5, fig=fig, rainbow=True)
            ax4.axis("off")
        set_labels(ax1, ax2, ax3, gene, number_of_genes, ress, n)

    sns.despine()
    fig.subplots_adjust(
        hspace=0.5,
        wspace=0.5,  # left=0.15, right=0.4, top=0.92, bottom=0.12
    )

    if save_plot:
        for ext in ["", ".png"]:
            fig.savefig(
                f"{rainbow_plot_path}{ext}",
                facecolor=fig.get_facecolor(),
                bbox_inches="tight",
                edgecolor="none",
                dpi=300,
            )
        plt.close(fig)
    return fig


def set_subfigure_titles(ax, n):
    if n == 0:
        ax[0].set_title("Rainbow plot", fontsize=7)
        ax[1].set_title("Phase portrait", fontsize=7)


def plot_gene_data(ax, ress, colors, add_line):
    plot_gene(ax[1], ress, colors, add_line)
    scatterplot(ax[0], ress, colors)


def scatter_gene(ax, adata, st, gene, basis):
    (index,) = np.where(adata.var_names == gene)
    return ax.scatter(
        adata.obsm[f"X_{basis}"][:, 0],
        adata.obsm[f"X_{basis}"][:, 1],
        s=3,
        c=st[:, index].flatten(),
        cmap="cividis",
    )


def adjust_subfigure(subfig):
    subfig[0].subplots_adjust(
        hspace=0.8, wspace=1.4, left=0.32, right=0.94, top=0.92, bottom=0.12
    )
    subfig[1].subplots_adjust(
        hspace=0.8, wspace=0.4, left=0.2, right=0.7, top=0.92, bottom=0.08
    )

    subfig[0].text(
        -0.025,
        0.58,
        "unspliced expression",
        size=7,
        rotation="vertical",
        va="center",
    )
    subfig[0].text(
        0.552,
        0.58,
        "spliced expression",
        size=7,
        rotation="vertical",
        va="center",
    )

    sns.despine()


@beartype
def get_genes(
    volcano_data: DataFrame,
    num_genes: int,
    negative_correlation: bool,
) -> Index:
    return (
        volcano_data.sort_values("mean_mae", ascending=False)
        .head(300)
        .sort_values("time_correlation", ascending=negative_correlation)
        .head(num_genes)
        .index
    )


def setup_scvelo_colors(adata, cell_state, basis):
    scv.pl.scatter(
        adata,
        basis=basis,
        fontsize=7,
        legend_loc="on data",
        legend_fontsize=7,
        color=cell_state,
        show=False,
    )
    return dict(
        zip(adata.obs.state_info.cat.categories, adata.uns["state_info_colors"])
    )


def setup_colors(adata, cell_state):
    clusters = adata.obs.loc[:, cell_state]
    return dict(
        zip(
            clusters.cat.categories,
            sns.color_palette("deep", clusters.cat.categories.shape[0]),
        )
    )


def get_posterior_samples_mean(data, posterior_samples):
    if (data[0] in posterior_samples) and (data[1] in posterior_samples):
        st = posterior_samples[data[0]].mean(0).squeeze()
        ut = posterior_samples[data[1]].mean(0).squeeze()
    else:
        st = posterior_samples["st_mean"]
        ut = posterior_samples["ut_mean"]
    return st, ut


def get_data(gene, st, ut, adata, cell_state, posterior_samples):
    (index,) = np.where(adata.var_names == gene)
    pos_mean_time = posterior_samples["cell_time"].mean(0).flatten()
    ress = pd.DataFrame(
        {
            "cell_time": pos_mean_time / pos_mean_time.max(),
            "cell_type": adata.obs[cell_state].values,
            "spliced": st[:, index].flatten(),
            "unspliced": ut[:, index].flatten(),
        }
    )
    return ress.sort_values("cell_time")


def plot_gene(ax1, ress, colors, add_line):
    sns.scatterplot(
        x="cell_time",
        y="spliced",
        data=ress,
        alpha=0.1,
        linewidth=0,
        edgecolor="none",
        hue="cell_type",
        palette=colors,
        ax=ax1,
        marker="o",
        legend=False,
        s=5,
    )
    if add_line:
        for row in range(ress.shape[0]):
            ax1.vlines(
                x=ress.cell_time[row],
                ymin=0,
                ymax=ress.spliced[row],
                colors=colors[ress.cell_type[row]],
                alpha=0.1,
            )


def set_labels(ax1, ax2, ax3, gene, ngenes, ress, n):
    if n == 0:
        ax3.set_title("Predictive spliced", fontsize=7)
    if n == ngenes - 1:
        ax1.set_xlabel("shared time", fontsize=7)
        ax1.set_ylabel("expression", fontsize=7)
        ax2.set_xlabel("spliced", fontsize=7)
        ax2.set_ylabel("uspliced", fontsize=7)
    else:
        ax1.set_xlabel("")
        ax2.set_xlabel("")
        ax1.set_ylabel("")
        ax2.set_ylabel("")
    ax2.text(
        -0.5,
        0.5,
        gene,
        fontsize=8,
        weight="normal",
        rotation=0,
        va="center",
        ha="right",
        transform=ax2.transAxes,
    )

    x_ticks = [0, np.power(10, get_closest_pow_of_10(ress["cell_time"].max()))]
    y_ticks = [0, np.power(10, get_closest_pow_of_10(ress["spliced"].max()))]
    ax1.set_xticks(x_ticks)
    ax1.set_yticks(y_ticks)
    ax1.get_xaxis().set_major_formatter(
        ticker.FuncFormatter(lambda x, _: construct_log_string(x))
    )
    ax1.get_yaxis().set_major_formatter(
        ticker.FuncFormatter(lambda x, _: construct_log_string(x))
    )
    ax1.tick_params(axis="both", which="major", labelsize=7)

    x_ticks = [0, np.power(10, get_closest_pow_of_10(ress["spliced"].max()))]
    y_ticks = [0, np.power(10, get_closest_pow_of_10(ress["unspliced"].max()))]
    ax2.set_xticks(x_ticks)
    ax2.set_yticks(y_ticks)
    ax2.get_xaxis().set_major_formatter(
        ticker.FuncFormatter(lambda x, _: construct_log_string(x))
    )
    ax2.get_yaxis().set_major_formatter(
        ticker.FuncFormatter(lambda x, _: construct_log_string(x))
    )
    ax2.tick_params(axis="both", which="major", labelsize=7)


def get_closest_pow_of_10(value):
    return np.floor(np.log10(value))


def construct_log_string(x):
    str_val = f"$10^{{{int(np.log10(x))}}}$" if x > 0 else "0"
    return rf"{str_val}"


def scatterplot(ax2, ress, colors):
    sns.scatterplot(
        x="spliced",
        y="unspliced",
        data=ress,
        alpha=0.4,
        linewidth=0,
        edgecolor="none",
        hue="cell_type",
        palette=colors,
        ax=ax2,
        marker="o",
        legend=False,
        s=3,
    )


def us_rainbowplot(
    genes: pd.Index,
    adata: AnnData,
    posterior_samples: Dict[str, ndarray],
    data: List[str] = ["st", "ut"],
    cell_state: str = "clusters",
) -> Figure:
    import matplotlib.lines as mlines

    fig, ax = plt.subplots(len(genes), 2)
    fig.set_size_inches(7, 14)
    n = 0
    if data[0] in posterior_samples:
        pos_s = posterior_samples[data[0]].mean(0).squeeze()
        pos_u = posterior_samples[data[1]].mean(0).squeeze()
    else:
        pos_u = posterior_samples["ut_mean"]
        pos_s = posterior_samples["st_mean"]

    for gene in genes:
        (index,) = np.where(adata.var_names == gene)
        ax1 = ax[n, 1]
        if n == 0:
            ax1.set_title("Rainbow plot")
        ress = pd.DataFrame(
            {
                "cell_time": posterior_samples["cell_time"].mean(0).squeeze(),
                "cell_type": adata.obs[cell_state].values,
                "spliced": pos_s[:, index].squeeze(),
                "unspliced": pos_u[:, index].squeeze(),
            }
        )
        if n == 2:
            sns.scatterplot(
                x="cell_time",
                y="spliced",
                data=ress,
                alpha=0.4,
                linewidth=0,
                edgecolor="none",
                hue="cell_type",
                ax=ax1,
                marker="o",
                legend="brief",
                palette="bright",
                s=10,
            )
        else:
            sns.scatterplot(
                x="cell_time",
                y="spliced",
                data=ress,
                alpha=0.4,
                linewidth=0,
                edgecolor="none",
                palette="bright",
                hue="cell_type",
                ax=ax1,
                marker="o",
                legend="brief",
                s=10,
            )

        ax2 = ax[n, 0]
        ax2.set_title(gene)
        ax2.set_ylabel("")
        ax2.set_xlabel("")
        sns.scatterplot(
            x="spliced",
            y="unspliced",
            data=ress,
            alpha=0.4,
            s=25,
            edgecolor="none",
            hue="cell_type",
            ax=ax2,
            legend=False,
            marker="*",
            palette="bright",
        )
        if n == 3:
            blue_star = mlines.Line2D(
                [],
                [],
                color="black",
                marker="o",
                linestyle="None",
                markersize=5,
                label="Spliced",
            )
            red_square = mlines.Line2D(
                [],
                [],
                color="black",
                marker="+",
                linestyle="None",
                markersize=5,
                label="Unspliced",
            )
            ax1.legend(
                handles=[blue_star, red_square], bbox_to_anchor=[2, -0.03]
            )
            ax1.set_xlabel("")
        n += 1
        ax1.legend(bbox_to_anchor=[2, 0.1])
        ax1.tick_params(labelbottom=True)
        ax2.set_xlabel("spliced")
        ax2.set_title(gene)

    plt.subplots_adjust(hspace=0.8, wspace=0.6, left=0.1, right=0.91)
    return fig
